# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Build Torch Addon."""

from __future__ import annotations

import argparse
import os
import shutil
import subprocess
import sys
import sysconfig
import tempfile
from collections.abc import Sequence
from pathlib import Path
# from ._system_path import system_include_dirs, system_library_dirs
try:
    import torch
    missing_dtypes = [
        "uint16", "uint32", "uint64",
        "float8_e5m2fnuz", "float8_e4m3fnuz",
        "float8_e4m3fn", "float8_e5m2"
    ]
    for dtype in missing_dtypes:
        if not hasattr(torch, dtype):
            setattr(torch, dtype, None)
except:
    pass

import torch.torch_version
import torch.utils.cpp_extension

import logging
logger = logging.getLogger(__name__)  # type: ignore


# Important: to avoid cyclic dependency, we avoid import tvm_ffi names at top level here.

IS_WINDOWS = sys.platform == "win32"
IS_DARWIN = sys.platform == "darwin"


def unique_paths(paths):
    seen = set()
    unique = []
    for p in paths:
        p_str = str(p)
        norm_p = os.path.abspath(p_str)
        if norm_p not in seen:
            seen.add(norm_p)
            unique.append(norm_p)  # 保留标准化后的路径
    return unique

def get_cpp_source(source_dir):
    source_path = []
    if isinstance(source_dir, list):
        for src_dir in source_dir:
            src = get_cpp_source(src_dir)
            source_path += src
    else:
        import glob
        source_dir = os.path.abspath(source_dir)
        for ext in ('*.cpp', '*.cc', '*.cxx'):
            source_path.extend(glob.glob(os.path.join(source_dir, '**', ext), recursive=True))

    return source_path

def get_torch_include_paths(build_with_cuda: bool) -> Sequence[str]:
    """Get the include paths for building with torch."""
    if torch.__version__ >= torch.torch_version.TorchVersion("2.6.0"):
        return torch.utils.cpp_extension.include_paths(
            device_type="cuda" if build_with_cuda else "cpu"
        )
    else:
        # type: ignore[call-arg]
        return torch.utils.cpp_extension.include_paths(cuda=build_with_cuda)

def get_cache_dir():
    return str(Path(os.environ.get("OMNIBACK_CACHE_DIR",
                            "~/.cache/omniback/")).expanduser())


def get_cache_lib(name: str, device: str, no_torch: bool):
    suffix = ".dll" if IS_WINDOWS else ".so"
    return os.path.join(get_cache_dir(), get_cache_name(name, device, no_torch))+suffix


def get_lib_name(name: str, device: str, no_torch: bool):
    suffix = ".dll" if IS_WINDOWS else ".so"
    return get_cache_name(name, device, no_torch)+suffix

def get_cache_name(name: str, device: str, no_torch: bool):
    # resolve library name
    if no_torch:
        return f"{name}-cxx11{torch.compiled_with_cxx11_abi()}"
    major, minor = torch.__version__.split(".")[:2]

    suffix = ".dll" if IS_WINDOWS else ".so"
    return f"{name}-torch{major}{minor}-{device}-cxx11{torch.compiled_with_cxx11_abi()}"

def main() -> None:  # noqa: PLR0912, PLR0915
    """Build the torch extension."""
    os.environ["TVM_FFI_DISABLE_TORCH_C_DLPACK"] = "1"
    from tvm_ffi.utils.lockfile import FileLock  # noqa: PLC0415

    parser = argparse.ArgumentParser(
        description="Build the torch c dlpack extension. After building, a shared library will be placed in the output directory.",
    )
    parser.add_argument(
        "--build-dir",
        type=str,
        required=False,
        help="Directory to store the built extension library. If not provided, a temporary directory will be used.",
    )
    parser.add_argument(
        '--source-dirs',
        type=str,
        nargs='+',  
        required=True,
        help='One or more source directories to search for C++ files'
    )
    parser.add_argument(
        '--include-dirs',
        type=str,
        nargs='+',
        required=True,
        help='One or more include directories to search for C++ headers'
    )
    parser.add_argument(
        '--ldflags',
        type=str,
        nargs='+',
        default=[],
        help='One or more include directories to search for C++ headers'
    )
    
    
    parser.add_argument(
        "--output-dir",
        type=str,
        required=False,
        default=get_cache_dir(),
        help="Directory to store the built extension library. If not specified, the default cache directory of omniback will be used.",
    )
    parser.add_argument(
        "--build-with-cuda",
        action="store_true",
        help="Build with CUDA support.",
    )
    parser.add_argument(
        "--no-torch",
        action="store_true",
        help="Build with no torch support.",
    )
    parser.add_argument(
        "--build-with-rocm",
        action="store_true",
        help="Build with ROCm support.",
    )
    parser.add_argument(
        "--name",
        type=str,
        required=True,
        help="The name of the generated library. ",
    )

    args = parser.parse_args()
    if args.build_with_cuda and args.build_with_rocm:
        raise ValueError("Cannot enable both CUDA and ROCm at the same time.")

    # resolve build directory
    if args.build_dir is None:
        build_dir = Path(tempfile.mkdtemp(prefix="omniback-torch"))
    else:
        build_dir = Path(args.build_dir)
    build_dir = build_dir.resolve()
    if not build_dir.exists():
        build_dir.mkdir(parents=True, exist_ok=True)

    if args.build_with_cuda:
        device = "cuda"
    elif args.build_with_rocm:
        device = "rocm"
    else:
        device = "cpu"
        
    libname = get_cache_name(args.name, device, args.no_torch)
    
    tmp_libname = libname + ".tmp"

    # create output directory is not exists
    output_dir = Path(args.output_dir).expanduser()
    if not output_dir.exists():
        output_dir.mkdir(parents=True, exist_ok=True)

    with FileLock(str(output_dir / (libname + ".lock"))):
        if (output_dir / libname).exists():
            # already built
            return

        # get the source
        source_dirs = [Path(d).expanduser() for d in args.source_dirs]
        ldflags = [str(d) for d in args.ldflags]
        source_path = get_cpp_source(source_dirs)
        if isinstance(args.include_dirs, str):
            args.include_dirs = [args.include_dirs]
        include_dirs = [Path(d).expanduser() for d in args.include_dirs]

        # resolve configs
        include_paths = [] + include_dirs
        cflags = []
        # include_paths.append(sysconfig.get_paths()["include"])

        # use CXX11 ABI
        if torch.compiled_with_cxx11_abi():
            cflags.append("-D_GLIBCXX_USE_CXX11_ABI=1")
        else:
            cflags.append("-D_GLIBCXX_USE_CXX11_ABI=0")
        
        if not args.no_torch:
            if args.build_with_cuda:
                cflags.append("-DBUILD_WITH_CUDA")
                if torch.utils.cpp_extension.CUDA_HOME is None:
                    logger.error("can not find CUDA_HOME")
            elif args.build_with_rocm:
                cflags.extend(torch.utils.cpp_extension.COMMON_HIP_FLAGS)
                cflags.append("-DBUILD_WITH_ROCM")
            include_paths.extend(get_torch_include_paths(
                args.build_with_cuda or args.build_with_rocm))

            for lib_dir in torch.utils.cpp_extension.library_paths():
                if IS_WINDOWS:
                    ldflags.append(f"/LIBPATH:{lib_dir}")
                else:
                    ldflags.extend(["-L", str(lib_dir)])
            
            import glob
            from torch.utils.cpp_extension import CUDA_HOME
            if not CUDA_HOME is None:
                cuda_lib_dir = os.path.join(CUDA_HOME, "lib64")
                # dirs = glob.glob(os.path.join(CUDA_HOME, "**/*/lib/stubs"))
                # assert len(dirs) == 1
                if IS_WINDOWS:
                    ldflags.append(f"/LIBPATH:{cuda_lib_dir}")
                else:
                    ldflags.extend(["-L", str(cuda_lib_dir)])

            # Add all required PyTorch libraries
            if IS_WINDOWS:
                # On Windows, use .lib format for linking
                ldflags.extend(
                    ["c10.lib", "torch.lib", "torch_cpu.lib"])
                if args.build_with_cuda:
                    ldflags.extend(["torch_cuda.lib", "c10_cuda.lib"])
            else:
                # On Unix/macOS, use -l format for linking
                ldflags.extend(
                    ["-lc10", "-ltorch", "-ltorch_cpu"])
                if args.build_with_cuda:
                    ldflags.extend(["-ltorch_cuda", "-lc10_cuda"])


        from omniback import get_include_dirs
        import omniback as om
        om_lib = om.libinfo.find_libomniback()
        
        ldflags.append(f"-L{os.path.dirname(om_lib)}")
        om_lib_name = os.path.splitext(os.path.basename(om_lib))[0].strip('lib')
        if om_lib_name.startswith('lib'):
            om_lib_name = om_lib_name[3:]
        ldflags.append(f"-l{om_lib_name}")
                
        from tvm_ffi.cpp.extension import build
        
        
    
        include_paths += get_include_dirs()
    
        include_paths = [str(x) for x in include_paths]
        include_paths = unique_paths(include_paths)
        result_lib = build(name=tmp_libname, cpp_files=[str(x) for x in source_path], extra_cflags=cflags,
                extra_ldflags=ldflags, extra_include_paths=include_paths, build_directory=build_dir)

        # rename the tmp file to final libname
        final_path=os.path.join(output_dir, os.path.basename(result_lib).replace(".tmp.", '.'))
        shutil.move(str(result_lib), final_path)
        print(f'saved to {final_path}')


# core
# python -m omniback.utils.build_lib --source-dirs csrc/torchplugins/ csrc/helper/ --include-dirs=csrc/ --build-with-cuda --name torchpipe_core

# nvjpeg
# python -m omniback.utils.build_lib --source-dirs csrc/nvjpeg_torch/ --include-dirs=csrc/ --build-with-cuda --ldflags="-lnvjpeg" --name torchpipe_nvjpeg

# opencv
# python -m omniback.utils.build_lib --no-torch --source-dirs csrc/mat_torch/ --include-dirs csrc/ /usr/local/include/opencv4/ --ldflags "-lopencv_core -lopencv_imgproc -lopencv_imgcodecs" --name torchpipe_opencv

if __name__ == "__main__":
    main()



import pytest
import omniback
from typing import Dict, Any

# import toml


from typing import Dict, Any


test_config = {
    "dag_base"
}
def test_configs():
    import time
    # time.sleep(8)
    for x in test_config:
        toml_path = f"config/{x}.toml"

        data = omniback.parse(toml_path)
        print(data)
        omniback.init("Interpreter", {}, omniback.Dict({"config": data}))


from enum import Enum
from typing import List, Optional

class Usage:
    def __init__(self) -> None:
        self.num_prompt_tokens: int = 0
        self.num_generated_tokens: int = 0
        self.num_total_tokens: int = 0

    def __repr__(self) -> str:
        return (f"Usage(num_prompt_tokens={self.num_prompt_tokens}, "
                f"num_generated_tokens={self.num_generated_tokens}, "
                f"num_total_tokens={self.num_total_tokens})")

class LogProbData:
    def __init__(self) -> None:
        self.token: str = ""
        self.token_id: int = 0
        self.logprob: float = 0.0
        self.finished_token: bool = False

    def __repr__(self) -> str:
        return (f"LogProbData(token={self.token}, token_id={self.token_id}, "
                f"logprob={self.logprob}, finished_token={self.finished_token})")

class LogProb:
    def __init__(self) -> None:
        self.token: str = ""
        self.token_id: int = 0
        self.logprob: float = 0.0
        self.finished_token: bool = False
        self.top_logprobs: Optional[List[LogProbData]] = None

    def __repr__(self) -> str:
        return (f"LogProb(token={self.token}, token_id={self.token_id}, "
                f"logprob={self.logprob}, finished_token={self.finished_token}, "
                f"top_logprobs={self.top_logprobs})")

class SequenceOutput:
    def __init__(self) -> None:
        self.index: int = 0
        self.text: str = ""
        self.token_ids: List[int] = []
        self.finish_reason: Optional[str] = None
        self.logprobs: Optional[List[LogProb]] = None

    def __repr__(self) -> str:
        return (f"SequenceOutput(index={self.index}, text={self.text}, "
                f"token_ids={self.token_ids}, finish_reason={self.finish_reason}, "
                f"logprobs={self.logprobs})")

class RequestOutput:
    def __init__(self) -> None:
        self.prompt: Optional[str] = None
        self.status: Optional['Status'] = None
        self.outputs: List[SequenceOutput] = []
        self.usage: Optional[Usage] = None
        self.finished: bool = False

    def __repr__(self) -> str:
        return (f"RequestOutput(prompt={self.prompt}, status={self.status}, "
                f"outputs={self.outputs}, usage={self.usage}, finished={self.finished})")

class StatusCode(Enum):
    OK = 0
    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    RESOURCE_EXHAUSTED = 5
    UNAUTHENTICATED = 6
    UNAVAILABLE = 7
    UNIMPLEMENTED = 8

class Status:
    def __init__(self, code: StatusCode, message: str) -> None:
        self._code = code
        self._message = message

    def __repr__(self) -> str:
        return f"Status(code={self._code}, message={self._message})"

    @property
    def code(self) -> StatusCode:
        return self._code

    @property
    def message(self) -> str:
        return self._message

    @property
    def ok(self) -> bool:
        return self._code == StatusCode.OK



from modules import ClipModule, UNetModule, VaeModule, SafetyModule
import modules
import sys
from helper import exp_config_parser
import toml
import json
import torch
import omniback
import torchpipe
from typing import List
import time
import random
import os
import math
from dataclasses import dataclass
import threading
import heapq
import numpy as np

import queue


class ThreadSafeDict:
    def __init__(self):
        self._dict = {}
        self._lock = threading.Lock()

    def get(self, key):
        with self._lock:
            return self._dict.get(key)

    def set(self, key, value):
        with self._lock:
            self._dict[key] = value

    def delete(self, key):
        with self._lock:
            if key in self._dict:
                del self._dict[key]


global_request_pool = {}
# NORM_TIME = 1
torch.set_grad_enabled(False)

# omniback.init("DebugLogger")
# with open('data/test_config.toml', 'r') as f:
#     data = toml.load(f)
# EXP_ID = os.getenv('EXP_ID')
# json_p = data[EXP_ID]['latency_profile']

# MBS = exp_config_parser.get_item('MBS', 0)
latency_profile = exp_config_parser.get_latency_profile()
# ref_times = exp_config_parser.get_ref_times()
# from dp import reorder_candidate_requests
torch.set_grad_enabled(False)

sys.path.insert(0, './')
assert os.environ["USE_TRT"] == "True"


# modules.USE_TRT = True


class StableDiffusionPipeline():
    def __init__(self, config_path, **kwargs):
        super().__init__()
        fp = open(config_path, "r")
        config = json.load(fp)
        self.stream_module_list = []
        print(config.keys())
        torch.set_grad_enabled(False)

        self.modules_type = {'ClipModule': ClipModule,
                             'UNetModule': UNetModule,
                             'VaeModule': VaeModule,
                             'SafetyModule': SafetyModule}
        self.modules = {}
        for instance_index in range(1):
            for key, value in config.items():
                if key != 'UNetModule':
                    continue
                value['instance_index'] = instance_index
                nk = key + f"{instance_index}"
                self.modules[nk] = self.modules_type[key](**value)
        for instance_index in range(1):
            for key, value in config.items():
                if key != 'UNetModule':
                    continue
                value['instance_index'] = instance_index
                nk = key + f"{instance_index}"
                self.modules[nk].deploy(**value)

        for key, value in config.items():
            if key == 'UNetModule':
                continue
            self.modules[key] = self.modules_type[key](**value)
            self.modules[key].deploy()

    def default_deploy(self, **kwargs):
        for module in self.stream_module_list:
            module.deploy()


# 初始化管道
sd_config_file = "stable_diffusion_v1_5/config.json"
sd_pipeline = StableDiffusionPipeline(config_path=sd_config_file)

sd_modules = sd_pipeline.modules


class Clip:
    def init(self, params, options):
        torch.set_grad_enabled(False)

    def forward(self, io: List[omniback.Dict]):
        # print(f'clip bs ={len(io)}')
        reqs = []
        for data in io:
            id = data['data']

            req = global_request_pool[id]
            # print('cl112p', type(id))
            data['result'] = id
            # data['stage'] = 1
            data['restart'] = 'continuous'
            # print('cli3p')
            reqs.append(req)

        sd_modules['ClipModule'].compute(reqs)

    def max(self):
        # print('xxxxxxxxxxxxxxd')
        return 16


class Unet:
    def init(self, params, options):
        torch.set_grad_enabled(False)
        print(f'unet params={params}')
        instance_index = int(params.pop('_independent_index'))
        self.module_name = f'UNetModule{instance_index}'
        self.sd_modules = sd_modules[self.module_name]

    def forward(self, io: List[omniback.Dict]):
        # print(f'unet bs = {len(io)}')
        # start_time = time.perf_counter()
        reqs = []
        for data in io:
            id = data['data']
            req = global_request_pool[id]
            for key in req.keys():
                if type(req[key]) == torch.Tensor and req[key].is_cpu:
                    req[key] = req[key].to('cuda')

            reqs.append(req)

            if req["loop_index"]["UNetModule"] + 1 < req['loop_num']["UNetModule"]:
                data['restart'] = 'continuous'

            elif req["loop_index"]["UNetModule"] + 1 == req['loop_num']["UNetModule"]:
                data['restart'] = 'continuous'
            data['result'] = id

            # print(req)
        self.sd_modules.compute(reqs)

        for req in reqs:
            req["loop_index"]["UNetModule"] += 1
        # print(io)
        # print(f'unet time = {time.perf_counter()-start_time}')

    def max(self):
        print('xxxxxxxxxxxxxxd4')
        return 10000


class Vae:
    def init(self, params, options):
        torch.set_grad_enabled(False)

    def forward(self, io: List[omniback.Dict]):
        # print('vae1')
        # print(f'vae bs ={len(io)}')
        # start_time = time.perf_counter()
        reqs = []
        for data in io:
            req_id = data['data']
            req = global_request_pool[req_id]
            data['result'] = req_id
            reqs.append(req)
            del global_request_pool[req_id]
        sd_modules['VaeModule'].compute(reqs)
        sd_modules['SafetyModule'].compute(reqs)

        # keys_to_remove = [k for k, v in req.items() if isinstance(v, torch.Tensor)]
        # for k in keys_to_remove:
        #     del req[k]
        for req in reqs:
            for k, v in req.items():
                if type(v) == torch.Tensor:
                    req[k] = None  # v.cpu()
                    # del req[k]
        # except Exception as e:
        #     print(f'error2 {e}', flush=True)
        # print(f'id v2: {id(req)}, {req_id} {req["id"]}')
        # print(f'vae time = {time.perf_counter()-start_time}')

    def max(self):
        print('xxxxxxxxxxxxxxd')
        return 32


@dataclass
class LoopInfo:
    data: omniback.Dict = None
    event: omniback.Event = None
    loop_index: int = 0
    loop_num: int = 1
    emergency: float = 0.0
    emergency_v2: float = 0.0
    delay: int = 0
    time: float = 0
    max_batch_size: int = 0
    stage: int = -1
    left_time: float = 0
    running: bool = False
    ratio_hardware: float = 0.0
    # start_iteration_time: float = 0
    # stop_iteration_time: float = 0

    # defination of emergency:
    # profile_post = (finish_time - stop_iteration_time)
    # SLO_iter = SLO - (start_iteration_time - request_time) - profile_post
    # left_time = SLO_iter - (now - start_iteration_time)
    # left_round = total_round - current_round
    # emergency = left_round/left_time


class PyContinuousBatching:
    def init(self, params, options):
        print(f'init foom {params}')
        self.receiving_data = {}
        self.lock = threading.Lock()
        self.input_queue = queue.Queue()
        self._stop_event = threading.Event()

        self.cached_data = {}
        # self.target = omniback.get(params.pop('target'))
        self.target = {}
        self.target[0] = omniback.get('node.clip_backend')
        self.target[1] = omniback.get('node.unet_backend')
        self.target[2] = omniback.get('node.vae_backend')

        # self.max_batch_size = int(params.pop('max_batch_size', 21+0))
        slo_factor = params.pop('slo_factor')
        print(f'params.pop(slo_factor) = {slo_factor}')
        slo_factor = [float(x) for x in slo_factor.split(',')]
        if len(slo_factor) == 2:
            slo_factor = slo_factor[0]
        else:
            slo_factor = slo_factor[0]
        self.slo_factor = float(slo_factor)

        from helper import exp_config_parser
        self.stage_thres = [4, 5, 4]
        self.config_parser = exp_config_parser.ConfigParser(step_delta=0.95)
        self.latency_profile = exp_config_parser.get_latency_profile()
        self.max_batch_size = self.config_parser.get_max_batch_size(
            int(self.slo_factor*10))
        # self.scheduler = self.config_parser.get_scheduler()
        self.instance_num = 1
        # if self.scheduler.startswith('instance_'):
        #     self.instance_num = int(self.scheduler.split('_')[1])

        # self.waiting_ids = set()
        print(
            f'PyContinuousBatching max_batch_size = {self.max_batch_size} {self.config_parser.get_slo_factor_to_batch_size()} slo_factor= {self.slo_factor}')
        self.delayed = set()

        # self.latency_ratio = ((self.latency_profile['safety'][8]+self.latency_profile['vae'][8])/8)/(self.latency_profile['unet'][10]/10 )

        print(self.latency_profile)
        self.AVG_h = self.latency_profile['clip'][1]
        self.AVG_t = self.latency_profile['safety'][1]+self.latency_profile['vae'][1]
        self.AVG_i = self.latency_profile['unet'][1]
        
        self.AVG_h = (self.AVG_h, self.AVG_h)
        self.AVG_t = (self.AVG_t, self.AVG_t)
        self.AVG_i = (self.AVG_i, self.AVG_i)
        
        print(
            f'AVG HEADER LATENCY = [{self.AVG_h}, AVG ITERATION LATENCY = {self.AVG_i}, AVG TAIL LATENCY = {self.AVG_t}')
        # exit(0)
        self.active_ids = []


        self.thread = threading.Thread(target=self.task_loop, daemon=True)
        self.thread.start()

        self.running_ids = set()

    def get_latency(self, bs, stage=2):
        if stage == 1:
            return self.latency_profile['unet'][bs]
        elif stage == 2:
            return (self.latency_profile['safety'][bs]+self.latency_profile['vae'][bs])
        else:
            return self.latency_profile['clip'][bs]

    def get_avg_latency(self, bs, stage=2):
        if stage == 2:
            return (self.latency_profile['safety'][bs]+self.latency_profile['vae'][bs])/bs

    def get_last_stage_avg_latency(self, num_lst_stage):
        last_stage_avg_latency = []
        if num_lst_stage >= self.stage_thres[2]:
            last_stage_avg_latency += [self.AVG_t[0]] * \
                ((num_lst_stage / self.stage_thres[2]) * self.stage_thres[2])
            bs = num_lst_stage % self.stage_thres[2]
            if bs != 0:
                last_stage_avg_latency += [self.get_avg_latency(bs)] * (bs)
        return last_stage_avg_latency

    def all_received(self):
        return all(local_id in self.receiving_data for local_id in self.running_ids)

    def __del__(self):
        self._stop_event.set()
        self.thread.join(timeout=5.0)  # 超时保护
        if self.thread.is_alive():
            print("Warning: Worker thread did not exit cleanly")

    def forward(self, ios: List[omniback.Dict]):
        self.input_queue.put(('forward', ios))

    def task_loop(self):
        while not self._stop_event.is_set():
            # all in one thread !!!! no race condition
            try:
                step, task = self.input_queue.get(timeout=0.1)
            except queue.Empty:
                continue
            self.on_start(task, step)

            #  if self.input_queue.empty():
            #     self.run_one_loop()

            self.input_queue.task_done()

    def on_start(self, ios, step):
        # print(f'on_start = {step}')
        if step == 'forward':
            for data in ios:
                # data['data'] is the request ID, and it's user's 责任去让他与实际数据一一对应
                self.receiving_data[data['data']] = data
        elif step == 'finish_stage':
            finished = self.on_finish_stage(ios)
            if len(finished) == 0:
                return
            for req_id in finished:
                del self.cached_data[req_id]
                self.running_ids.discard(req_id)
                print(f'request {req_id} finished')
        else:
            assert False, f'error step {step}'

        # print('receiving_data|running_ids: ', self.receiving_data.keys(), self.running_ids)
        if not self.input_queue.empty() or not self.all_received():
            return
        else:
            receiving_data = self.receiving_data
            self.receiving_data = {}
            self.running_ids = set()
        # self.num_stages = {i:0 for i in range(3)}
        # print('all data received: ', list(receiving_data.keys()))

        # start_scheduler_time = time.time()
        # all data received. update receiving_data to cached_data (receiving_data >= cached).
        # first_round_data = []
        for req_id, data in receiving_data.items():
            # print(f'parsing {req_id}')
            req = global_request_pool[req_id]  # only for debug

            # import pdb;pdb.set_trace()
            loop_num = data['loop_num']
            if not req_id in self.cached_data:
                # assert req['loop_index']['UNetModule'] == 0
                # assert req['loop_num']['UNetModule'] == loop_num
                req['start_iteration_time'] = time.time()
                ev = data.pop('event')
                self.cached_data[req_id] = LoopInfo(
                    data, ev, 0, loop_num=loop_num, time=req['start_iteration_time'], stage=0)
                # self.num_stages[0] += 1
            else:
                ev = data.pop('event')
                v = self.cached_data[req_id]
                if v.stage == 0:
                    v.stage += 1
                    v.loop_index = 0
                else:
                    v.loop_index += 1
                    if v.loop_index == v.loop_num:
                        v.stage += 1

                v.data = data
                v.event = ev
                # v.stage = stage
                v.loop_num = loop_num  # can be changed per iteration
                # self.num_stages[v.stage] += 1
                # print(f'on start id={req_id} stage={v.stage}, loop_index={v.loop_index}')

        # clear data
        # print(receiving_data, self.cached_data)
        receiving_data.clear()

        self.update_resource_status()
        if len(self.active_ids) == 0:
            return

        # activate_status = [(self.cached_data[x].stage, int(100*self.cached_data[x].ratio_hardware)) for x in active_ids]

        # print(f'activate ids: {active_ids} status={activate_status}')

        stage, most_emergent_ids, largest_status = self.get_most_emergent_ids(
            self.active_ids)
        # unactivate_ids = [x for x in active_ids if x not in most_emergent_ids]

        # self.active_ids = most_emergent_ids
        # print(f'decision => stage = {stage}, ids = {most_emergent_ids}/{unactivate_ids}, largest_status={largest_status}')
        self.running_ids = set(most_emergent_ids)

        self.run_stage(most_emergent_ids, stage)

        # all_time = time.time() - start_scheduler_time
        # print(f'scheduler time = {all_time}, cached = {len(self.cached_data)}, running = {len(self.running_ids)}, active_ids = {len(active_ids)}')
        # # exit(0)

    def get_most_emergent_ids(self, active_ids):
        largest_id = max(
            active_ids, key=lambda x: self.cached_data[x].ratio_hardware)
        stage = self.cached_data[largest_id].stage
        left_time = self.cached_data[largest_id].left_time
        ori_active_ids = [
            x for x in active_ids if self.cached_data[x].stage == stage]
        ori_active_ids.sort(
            key=lambda x: self.cached_data[x].ratio_hardware, reverse=True)
        # bs = 1
        # while (self.get_latency(bs +1, stage) < left_time) and bs +1 <= len(active_ids):
        #     bs += 1
        hr = self.cached_data[largest_id].ratio_hardware
        bs = min(int(1/hr), len(ori_active_ids))
        # bs = min(self.max_batch_size, len(ori_active_ids))
        # if bs >10:
        #     bs = int(bs/2)
        # elif bs > 10:
        #     bs = int(bs/2 + 1)
        # elif bs >= 8:
        #     bs = 7
        # bs = min(bs, self.stage_thres[stage] * 2)
        active_ids = ori_active_ids[:bs]
        # print(f'active_ids = {active_ids},stage={stage} hr={hr:.3f} act={bs}/{len(ori_active_ids)} total={len(self.cached_data)}')
        # (largest_id, self.cached_data[largest_id].ratio_hardware, left_time, bs)
        return stage, active_ids, None

    def update_resource_status(self):
        need_drop = set()

        candidate_requests = []

        # num_stage_last = []
        num_stage_last = sum(
            1 for x in self.cached_data.values() if x.stage == 2)

        for req_id, v in self.cached_data.items():  # 修改循环变量为v
            req = global_request_pool[req_id]
            SLO = req['SLO']
            request_time = req['request_time']
            now = time.time()
            v.left_time = max(1e-6, request_time + SLO - now)

            if v.stage == 0:
                v.resource = (
                    self.AVG_h[1] + self.AVG_t[1] + v.loop_num * self.AVG_i[0])

                if (self.AVG_h[1] + self.AVG_t[1] + (v.loop_num) * self.AVG_i[1]) > v.left_time:
                    v.resource += v.left_time
            elif v.stage == 1:
                v.resource = (
                    self.AVG_t[1] + (v.loop_num - v.loop_index) * self.AVG_i[0])
                # if v.resource / v.left_time > 0.2:
                #     v.resource += v.left_time * 3
                # if (v.loop_num - v.loop_index) * self.AVG_i[0]/ (v.left_time - self.AVG_t[1]) > 0.2:
                #     resource += v.left_time * 3
                if (self.AVG_t[1] + (v.loop_num - v.loop_index) * self.AVG_i[1]) > v.left_time:
                    v.resource += v.left_time*2
                # if resource > v.left_time * 0.5:
                #     resource = v.left_time * 1.5
            elif v.stage == 2:
                v.resource = self.AVG_t[1]
            else:
                assert False, v.stage

            v.ratio_hardware = v.resource / v.left_time

            if v.ratio_hardware >= 1:
                need_drop.add(req_id)
            else:
                if v.stage != 2:
                    candidate_requests.append(
                        (req_id, v.resource, v.left_time))
                elif num_stage_last > 0:
                    # 修正最后一阶段的资源占用情况
                    v.resource = (self.latency_profile['safety'][num_stage_last] +
                                  self.latency_profile['vae'][num_stage_last])/num_stage_last
                    # v.ratio_hardware = resource / v.left_time
                    # resource += 1e-6 *(1- v.ratio_hardware)
                    candidate_requests.append(
                        (req_id, v.resource, v.left_time))
        # drop2, candidate_requests = reorder_candidate_requests(candidate_requests)

        # print(f'candidate_requests = {candidate_requests}')

        # last_stage_avg_latency = self.get_last_stage_avg_latency(len(num_stage_last))

        # need_drop = need_drop.union(drop2)

        if len(need_drop) > 0:
            # print(f'drop2 = {drop2}')
            self.drop_request(need_drop)

        self.active_ids = [x for x in self.active_ids if x in self.cached_data]
        sum_resource = sum(
            [self.cached_data[x].resource for x in self.active_ids])
        if sum_resource >= 1:
            # candidate = self.active_ids
            pass
        else:
            candidate_requests = [
                x[0] for x in candidate_requests if x[0] not in self.active_ids]
            # candidate_requests.sort(key=lambda x: (self.cached_data[x].resource))
            candidate_requests.sort(key=lambda x: (
                self.cached_data[x].delay, self.cached_data[x].time))
            # z = [ (x, self.cached_data[x].delay, self.cached_data[x].resource) for x in candidate_requests]
            # if len(z) > 0:
            #     print(z)
            self.active_ids += candidate_requests
        # sum_rsc =
        # for req_id in self.active_ids:

        cumulative_ratio = 0
        active_ids = []  # {i:[] for i in range(3)}
        for req_id in self.active_ids:
            v = self.cached_data[req_id]
            if cumulative_ratio + v.ratio_hardware < 1.0:
                cumulative_ratio += v.ratio_hardware
                active_ids.append(req_id)
        self.active_ids = active_ids

        for k, v in self.cached_data.items():
            if k not in self.active_ids:
                v.delay += 1
                # active_ids[v.stage].append(req_id)
        # print([(x, self.cached_data[x].ratio_hardware) for x in self.active_ids])
        # 总的资源量越小越好

    def run_stage(self, ids, target):
        ios = []
        for k in ids:
            v = self.cached_data[k]
            ios.append(v.data)
            assert 'event' not in v.data

        ev = omniback.Event(len(ios))
        ev.set_callback(lambda: self.input_queue.put(('finish_stage', ids)))
        ev.set_exception_callback(lambda x: print(
            f'error {x} in Stage {target} '))
        for item in ios:
            item['event'] = ev
        # print(f'cb forward {len(io)}')
        self.target[target].forward(ios, None)

    def on_finish_stage(self, ids):
        should_del = []
        for k in ids:
            v = self.cached_data[k]
            if v.stage == 0 or v.stage == 1:
                # v.stage = 1
                # print(f'request {k} in stage {v.stage} finished')
                pass

                # if v.loop_index + 1 >=v.loop_num:
                #     v.stage += 1
                # req['stop_iteration_time'] = stop_iteration_time
                # req['stop_iteration_batch'] = len(self.cached_data)
            elif v.stage == 2:
                should_del.append(k)

            v.data['event'] = v.event
            v.event.notify_all()
            # print(f'request {k} notified')

        return should_del
    # def on_finish(self, ids):
    #     should_del = []
    #     for k in ids :
    #         v=  self.cached_data[k]
    #         if v.loop_index + 1 >=v.loop_num:
    #             should_del.append(k)
    #             req = global_request_pool[k]
    #             # req['stop_iteration_time'] = stop_iteration_time
    #             # req['stop_iteration_batch'] = len(self.cached_data)
    #         v.data['event'] = v.event
    #         v.event.notify_all()
    #             # print(f'request {k} notified')
    #     for req_id in should_del:
    #         del self.cached_data[req_id]
    #         print(f'request {req_id} finished')
        # with self.lock:
        #     self.start_batching = True

    # def ff(self, x):
    #     # e = self.cached_data[x].emergency
    #     # thres = 30
    #     left = self.cached_data[x].loop_num - self.cached_data[x].loop_index
    #     return left# * (1- thres / e)

    def ff(self, x):
        # return (x[1].delay, x[1].emergency)
        return (x[1].delay, x[1].time)
        # left / (time - x / thr) <= thr
        # =>
        # e = x[1].emergency
        # thres = 1/(self.time_one_loop +1e-8)
        # print(thres)
        left = x[1].loop_num - x[1].loop_index
        return left  # * (thres/min(e,thres) - 1)

    def delay_some_request(self):
        # old_delayed = self.delayed
        # thres = 8
        delayed = set()
        need_delay = len(self.cached_data) - self.max_batch_size
        if need_delay <= 0:
            return delayed
        # pool = set(x for x  in self.cached_data.keys() if x not in  old_delayed)
        # if need_delay <= len(old_delayed):
        #     pass
        # else:

        #     for req_id in heapq.nsmallest(need_delay-len(old_delayed), pool, key=lambda x: self.cached_data[x].emergency):

        # need_drop = 0
        # for req_id, v in heapq.nsmallest(need_delay, self.cached_data.items(), key=lambda x: x[1].emergency):
        #     if v.emergency > thres:
        #         need_drop += 1

        for req_id, v in heapq.nlargest(need_delay, self.cached_data.items(), key=self.ff):
            v.delay += 1
            delayed.add(req_id)
            # print(f'delay: id={req_id} times={v.delay}, MBS={self.max_batch_size} em={v.emergency}')

        return delayed

    def drop_request(self, req_ids):
        for req_id in req_ids:
            v = self.cached_data[req_id]
            left_num = v.loop_num - v.loop_index
            print(
                f'drop req. {req_id}. hardware ratio = {v.ratio_hardware} left_time = {v.left_time}, left_num={left_num}')

            global_request_pool[req_id]['loop_index']['UNetModule'] = -1

            v.data['event'] = v.event
            v.event.notify_all()

            del self.cached_data[req_id]
            del global_request_pool[req_id]

        return


omniback.register("Clip", Clip)
omniback.register("Unet", Unet)
omniback.register("Vae", Vae)
omniback.register("PyContinuousBatching", PyContinuousBatching)


def get_scheduler(slo_factor):
    # omniback.pipe({'unet_backend':{'backend':'S[unet, SyncTensor,TimeStamp(finish_iter)]','batching_timeout':1,'instance_num':1},
    #     })
    omniback.pipe({'unet_backend': {'backend': 'SyncTensor[Unet]', 'batching_timeout': 0, 'instance_num': 1},
               'clip_backend': {'backend': 'SyncTensor[Clip]', 'batching_timeout': 8, 'instance_num': 1},
               'vae_backend': {"backend": 'SyncTensor[Vae]', 'batching_timeout': 8, 'instance_num': 1}
               })

    config = {
        'global': {'entrypoint': 'Restart[DagDispatcher]'},
        # 'clip':{'backend':'SyncTensor[Clip]','batching_timeout':8,'instance_num':1,'next':'unet'},  #
        'continuous': {'node_entrypoint': 'Register[PyContinuousBatching]', 'slo_factor': slo_factor}}

    scheduler = omniback.pipe(config)

    return scheduler


class Model:
    def __init__(self, slo_factor):
        # omniback.init('DebugLogger')
        if isinstance(slo_factor, tuple):
            assert len(slo_factor) == 2
            slo_factor = str(slo_factor[0]) + ','+str(slo_factor[1])
        self.scheduler = get_scheduler(slo_factor)

    def __call__(self, req_id, request, exc_cb, call_back):
        event = omniback.Event()
        # print('before set_exception_callback')
        event.set_exception_callback(exc_cb)
        # print('after set_exception_callback')

        def drop_cb():
            if request['loop_index']['UNetModule'] != request['loop_num']['UNetModule']:
                # print(f'drop_cb, id={request["id"]} {id(request)}')
                for k, v in request.items():
                    if k not in ['request_time', 'loop_index', 'loop_num', 'exception', 'id', "SLO"]:
                        request[k] = None
                # print(request)

        event.set_callback(call_back)
        event.set_callback(drop_cb)
        # print(f'id={req_id}')

        io = omniback.Dict({'data': req_id, 'node_name': 'continuous', 'event': event,
                       'stage': 0, 'loop_num': request['loop_num']["UNetModule"]})
        # io.pop('event', None)
        # print(io)
        assert req_id not in global_request_pool
        global_request_pool[req_id] = request

        # start = time.perf_counter()
        self.scheduler(io)
        # event.wait()
        # print(f'total = {time.perf_counter() - start}')


def main(save_img=False, num_request=100, image_size=256):
    prompts = [
        "a photograph of an astronaut riding a horse",
        "A cute otter in a rainbow whirlpool holding shells, watercolor",
        "An avocado armchair",
        "A white dog wearing sunglasses"
    ]
    prompt = random.choice(prompts)
    id = 0

    scheduler = get_scheduler()

    def run():
        request = {
            "prompt": random.choice(prompts),
            "height": image_size,
            "width": image_size,
            "loop_num": {"UNetModule": random.randint(30, 50)},
            "request_time": time.time(),
            "guidance_scale": 7.5,
            "seed": 42 + id,  # 每轮使用不同的种子
            "SLO": 10000,
            "loop_index": {"UNetModule": 0},
            "id": id
        }
        global_request_pool[id] = request

        io = {'data': id, 'node_name': 'continuous',
              'loop_num': request['loop_num']["UNetModule"]}
        scheduler(io)
        print(
            f"(id={request['id']}) Server-side end2end latency: {time.perf_counter() - request['request_time']}")

        return request

    for index in range(20):
        run()  # warm up
        print(f'warm up {index}')
    print('warm up finished')

    start = time.perf_counter()

    for _ in range(num_request):
        run()
    total_time = (time.perf_counter() - start)/num_request
    print(f"time: {total_time:.4f} second")

    # print(global_request_pool[id])
    if save_img:
        request = run()
        print(request.keys())
        img = request['pillow_image']
        img.save(f"output_image_run{id}.jpg", quality=95)
        print(f"saved to output_image_run{id}.jpg")


if __name__ == "__main__":
    main(save_img=True, num_request=100, image_size=256)



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import matplotlib
from matplotlib.lines import Line2D
from matplotlib.colors import LinearSegmentedColormap

# === 增强的全局设置 ===
matplotlib.rcParams.update({
    'font.size': 14,
    'axes.titlesize': 16,
    'axes.labelsize': 14,
    'xtick.labelsize': 12,
    'ytick.labelsize': 12,
    'legend.fontsize': 12,
    'figure.dpi': 600,
    'figure.figsize': (10, 6.5),
    'axes.grid': True,
    'grid.linestyle': '--',
    'grid.alpha': 0.3
})

# 读取数据
df = pd.read_csv("benchmark_results.csv")

# 计算基准吞吐量
base_throughputs = {}
for model in df['model'].unique():
    base = df[(df['model'] == model) & (
        df['batch_size'] == 64)]['throughput'].values
    if len(base) > 0:
        base_throughputs[model] = base[0]

# 计算归一化吞吐量
df['normalized_throughput'] = df.apply(
    lambda row: row['throughput'] / base_throughputs[row['model']
                                                     ] if row['model'] in base_throughputs else 1.0,
    axis=1
)

# 过滤数据
mbs = 13
df = df[df['batch_size'] <= mbs]

# 模型显示名称处理


def get_display_name(orig_name):
    if orig_name.endswith('.py'):
        target = orig_name[:-3]
        if target == 'seg':
            return 'Segmentation Postprocessing', True
        if target == 'jpg_decode':
            return 'JPEG Decoding', True
        return target.replace('_', ' ').title(), True
    return orig_name, False


# 模型分组和排序
models = df['model'].unique().tolist()
models.sort(key=lambda model: (not model.endswith('.py'), model))
model_ids = {model: idx for idx, model in enumerate(models)}
df['model_id'] = df['model'].map(model_ids)

# === 增强视觉设计 ===
# 使用高对比度、高饱和度的配色方案
high_contrast_colors = [
    '#e6194b', '#3cb44b', '#4363d8', '#f58231', '#911eb4',
    '#46f0f0', '#f032e6', '#bcf60c', '#fabebe', '#008080',
    '#e6beff', '#9a6324', '#800000', '#aaffc3', '#808000'
]

# 定义多种标记符号
markers = ['o', 's', '^', 'D', 'v', '<', '>', 'p', 'h', 'd', '*', 'X', 'P']

# 线型设置
line_styles = {
    'image_processing': (0, (4, 1.5)),  # 更明显的虚线
    'model_inference': (0, (1, 0))       # 实线
}

# 创建图形
fig, ax = plt.subplots(figsize=(8, 5))

# 存储图例元素
model_legend_elements = []
task_legend_elements = []

# 绘制曲线
for orig_name in models:
    display_name, is_image = get_display_name(orig_name)
    model_df = df[df['model'] == orig_name].sort_values('batch_size')

    # 确定样式
    task_type = 'image_processing' if is_image else 'model_inference'
    color_idx = model_ids[orig_name] % len(high_contrast_colors)
    base_color = high_contrast_colors[color_idx]

    # 分配视觉属性
    line_style = line_styles[task_type]
    marker = markers[color_idx % len(markers)]
    line_width = 2.5 if is_image else 2.0  # 图像处理任务使用更粗的线条

    # 绘制曲线
    ax.plot(
        model_df['batch_size'],
        model_df['normalized_throughput'],
        marker=marker,
        markersize=8,
        markeredgewidth=2.0,
        markeredgecolor=base_color,
        markerfacecolor=base_color if is_image else 'white',
        linestyle=line_style,
        linewidth=line_width,
        color=base_color,
        alpha=0.95,
        zorder=10 if is_image else 5,
        markevery=(mbs-1, 1)
    )

    # 添加模型图例元素
    model_legend_elements.append(Line2D(
        [0], [0],
        marker=marker,
        color=base_color,
        markeredgecolor=base_color,
        markerfacecolor=base_color if is_image else 'white',
        markersize=10,
        markeredgewidth=1.5,
        linestyle='',
        label=display_name
    ))

    # 标记75%性能点
    for idx, row in model_df.iterrows():
        if row['normalized_throughput'] >= 0.75:
            ax.plot(
                row['batch_size'],
                row['normalized_throughput'],
                marker='*',
                markersize=18,
                markeredgewidth=2.0,
                markerfacecolor=base_color,#'gold',
                markeredgecolor=base_color,  # 修改这里：将 'black' 改为 base_color
                zorder=20
            )
            break

# === 新增：计算参考点性能对比 ===
reference_point_results = []

for orig_name in models:
    model_df = df[df['model'] == orig_name].sort_values('batch_size')

    # 找到第一个达到75%性能的点
    ref_point = None
    for idx, row in model_df.iterrows():
        if row['normalized_throughput'] >= 0.75:
            ref_point = row
            break

    if ref_point is not None:
        ref_batch_size = ref_point['batch_size']
        ref_throughput = ref_point['throughput']

        # 获取batch_size=1的数据
        bs1_data = model_df[model_df['batch_size'] == 1]
        if not bs1_data.empty:
            bs1_throughput = bs1_data['throughput'].values[0]

            # 计算吞吐量提升倍数
            throughput_ratio = ref_throughput / bs1_throughput

            # 计算延迟降低倍数（吞吐量的倒数）
            latency_ratio = (bs1_throughput) / (ref_throughput/ref_batch_size)

            display_name, is_image = get_display_name(orig_name)

            reference_point_results.append({
                'model': display_name,
                'reference_batch_size': ref_batch_size,
                'throughput_ratio': throughput_ratio,
                'latency_ratio': latency_ratio,
                'is_image_processing': is_image
            })

print("=" * 80)
print("参考点性能对比分析 (与batch_size=1相比)")
print("=" * 80)
print(f"{'模型':<25} {'参考点Batch Size':<18} {'吞吐倍数':<15} {'延迟倍数':<15}")
print("-" * 80)

for result in reference_point_results:
    print(f"{result['model']:<25} {result['reference_batch_size']:<18} "
          f"{result['throughput_ratio']:<15.2f} {result['latency_ratio']:<15.2f}")

avg_throughput_ratio = np.mean([r['throughput_ratio']
                               for r in reference_point_results])
avg_latency_ratio = np.mean([r['latency_ratio']
                            for r in reference_point_results])

print("-" * 80)
print(f"{'平均值':<25} {'-':<18} {avg_throughput_ratio:<15.2f} {avg_latency_ratio:<15.2f}")
print("=" * 80)


ax.axhline(y=0.75, color='#333333', linestyle=':', linewidth=3.0, alpha=0.9)
ax.text(0.91, 0.594, '75% Reference',  # 0.63
        transform=ax.transAxes, ha='right', fontsize=12, color='#333333')

ax.set_xlabel('Batch Size', fontsize=15, labelpad=10)
ax.set_ylabel('Normalized Offline Throughput', fontsize=15, labelpad=10)
ax.set_xticks(range(1, mbs+1))
ax.set_xlim(0.8, mbs+0.2)
ax.set_ylim(0.1, 1.1)
ax.set_yticks(np.arange(0.2, 1.1, 0.2))
ax.grid(True, which='major', linestyle='--', alpha=0.4)

task_legend_elements.append(
    Line2D([0], [0], color='black', lw=2.5, linestyle=line_styles['model_inference'], label='Model Inference'))
task_legend_elements.append(Line2D(
    [0], [0], color='black', lw=2.5, linestyle=line_styles['image_processing'], label='Image Processing'))
task_legend_elements.append(Line2D([0], [0], marker='*', color='w', markerfacecolor='black',
                                   markersize=16, markeredgecolor='black', label='First ≥75% Point'))

task_legend = ax.legend(
    handles=task_legend_elements,
    loc='upper left',
    framealpha=0.75,
    facecolor='white',
    frameon=True,
    edgecolor='lightgray',
    title='',
    title_fontsize=13
)

model_legend = ax.legend(
    handles=model_legend_elements,
    loc='lower right',
    framealpha=0.95,
    facecolor='white',
    ncol=1,
    frameon=True,
    edgecolor='lightgray',
    title='',
    title_fontsize=13
)

ax.add_artist(model_legend)
ax.add_artist(task_legend)

plt.tight_layout(rect=[0, 0, 1, 0.95])

plt.savefig('normalized_throughput.pdf',
            bbox_inches='tight',
            dpi=600,
            pad_inches=0.1)
print(f'saved to normalized_throughput.pdf')



import subprocess
from pathlib import Path
from omniback.utils import build_lib
from .utils._cache_setting import get_cache_dir

import ctypes, os, sys
import logging


logger = logging.getLogger(__name__)  # type: ignore

csrc_dir = os.path.dirname(__file__)

def load_whl_lib(path_of_cache):
    p = os.path.join(os.path.dirname(__file__), 'lib',
                     os.path.basename(path_of_cache))
    if os.path.exists(p):
        ctypes.CDLL(p, mode=ctypes.RTLD_GLOBAL)
        logger.info(f'Successfully loaded precompiled {p} from the installed package')
        return True
    return False

def get_whl_lib(path_of_cache):
    p = os.path.join(os.path.dirname(__file__), 'lib',
                     os.path.basename(path_of_cache))
    if os.path.exists(p):
        # ctypes.CDLL(p, mode=ctypes.RTLD_GLOBAL)
        # logger.info(f'Successfully loaded precompiled {p} from the installed package')
        return p
    return None

def _load_lib_with_torch_cuda(name):
    # import torch
    # device = f"cuda{torch.version.cuda.split('.')[0]}"
    device = "cuda"
    local_lib = build_lib.get_cache_lib(
        name, device, False)
    if load_whl_lib(local_lib):
        return True
    if name == "torchpipe_tensorrt":
        if os.path.exists(local_lib):
            try:
                ctypes.CDLL(local_lib, mode=ctypes.RTLD_GLOBAL)
            except OSError:
                from .utils._build_trt import get_trt_include_lib_dir
                _, lib_dir = get_trt_include_lib_dir()
                if lib_dir is None:
                    import torch
                    cuda_version = int(torch.version.cuda.split('.')[0])

                    logger.warning(
                        f"Can not find TensorRT. Skip load torchpipe_tensorrt. Set TENSORRT_INCLUDE and TENSORRT_LIB")
                    #  or `pip install tensorrt-cu{cuda_version}`
                    has_import_trt=False
                    try:
                        import tensorrt
                        has_import_trt = True
                    except:
                        return False
                    if has_import_trt:
                        ctypes.CDLL(local_lib, mode=ctypes.RTLD_GLOBAL)
                        return True
                        
                os.environ["LD_LIBRARY_PATH"] = f"{lib_dir}:" + \
                    os.environ.get("LD_LIBRARY_PATH", "")
                nvinfer = Path(lib_dir)/"libnvinfer.so"
                nvonnxparser = Path(lib_dir)/"libnvonnxparser.so"
                nvinfer_plugin = Path(lib_dir)/"libnvinfer_plugin.so"

                ctypes.CDLL(nvinfer.resolve(), mode=ctypes.RTLD_GLOBAL)
                ctypes.CDLL(nvonnxparser.resolve(), mode=ctypes.RTLD_GLOBAL)
                ctypes.CDLL(nvinfer_plugin.resolve(), mode=ctypes.RTLD_GLOBAL)
                
                ctypes.CDLL(local_lib, mode=ctypes.RTLD_GLOBAL)
            return True
    else:
        if os.path.exists(local_lib):
            ctypes.CDLL(local_lib, mode=ctypes.RTLD_GLOBAL)
            return True
    return False

def _load_lib(name):
    if name == "torchpipe_opencv":
        torchpipe_opencv = build_lib.get_cache_lib(
            "torchpipe_opencv", "", True)
        if load_whl_lib(torchpipe_opencv):
            return True
        if os.path.exists(torchpipe_opencv):
            try:
                ctypes.CDLL(torchpipe_opencv, mode=ctypes.RTLD_GLOBAL)
            except OSError:
                from .utils._build_cv import get_cv_include_lib_dir, is_system_exists_cv, get_system_cv
                _, lib_dir = get_cv_include_lib_dir()
                if lib_dir is None:
                    if is_system_exists_cv():
                        _, lib_dir = get_system_cv()
                        
                if lib_dir is None:
                    raise RuntimeError(
                        "can not find opencv library. You can set it through OPENCV_LIB")
                    
                os.environ["LD_LIBRARY_PATH"] = f"{lib_dir}:" + \
                    os.environ.get("LD_LIBRARY_PATH", "")
                
                core = Path(lib_dir)/"libopencv_core.so"
                imgproc = Path(lib_dir)/"libopencv_imgproc.so"
                imgcodecs = Path(lib_dir)/"libopencv_imgcodecs.so"
                
                ctypes.CDLL(core.resolve(), mode=ctypes.RTLD_GLOBAL)
                ctypes.CDLL(imgproc.resolve(), mode=ctypes.RTLD_GLOBAL)
                ctypes.CDLL(imgcodecs.resolve(), mode=ctypes.RTLD_GLOBAL)
                    
                ctypes.CDLL(torchpipe_opencv, mode=ctypes.RTLD_GLOBAL)
            return True
    else:
        return _load_lib_with_torch_cuda(name)
    return False




    
def _build_lib(name):
    logger.warning(
        f'Pre-built library not found for {name}, starting JIT compilation')
    if name == "torchpipe_core":
        # python -m omniback.utils.build_lib --source-dirs csrc/torchplugins/ csrc/helper/ --include-dirs=csrc/ --build-with-cuda --name torchpipe_core
        subprocess.run(
            [
                sys.executable,
                "-m",
                "omniback.utils.build_lib",
                "--source-dirs",
                os.path.join(csrc_dir, "csrc/torchplugins/"),
                os.path.join(csrc_dir, "csrc/helper/"),
                "--include-dirs",
                os.path.join(csrc_dir, "csrc/"),
                "--build-with-cuda",
                "--name",
                "torchpipe_core"
            ],
            check=True,
            env={**os.environ, "EXAMPLE_ENV": "1"},
        )
    elif name == "torchpipe_nvjpeg":
        # python -m omniback.utils.build_lib --source-dirs csrc/nvjpeg_torch/ --include-dirs=csrc/ --build-with-cuda --ldflags="-lnvjpeg" --name torchpipe_nvjpeg
        subprocess.run(
            [
                sys.executable,
                "-m",
                "omniback.utils.build_lib",
                "--source-dirs",
                os.path.join(csrc_dir, "csrc/nvjpeg_torch/"),
                "--include-dirs",
                os.path.join(csrc_dir, "csrc/"),
                "--build-with-cuda",
                "--ldflags=-lnvjpeg",
                "--name",
                "torchpipe_nvjpeg"
            ],
            check=True,
            # TVM_FFI_DISABLE_TORCH_C_DLPACK
            env={**os.environ, "EXAMPLE_ENV": "1"},
        )
    elif name == "torchpipe_tensorrt":
        # python -m omniback.utils.build_lib --source-dirs csrc/tensorrt_torch/ --include-dirs=csrc/ --build-with-cuda --ldflags="-lnvinfer -lnvonnxparser  -lnvinfer_plugin" --name torchpipe_tensorrt
        from .utils._build_trt import _build_trt
        _build_trt(csrc_dir)
 
    elif name == "torchpipe_opencv":
        # python -m omniback.utils.build_lib --no-torch --source-dirs csrc/mat_torch/ --include-dirs csrc/ /usr/local/include/opencv4/ --ldflags "-lopencv_core -lopencv_imgproc -lopencv_imgcodecs" --name torchpipe_opencv
        from .utils._build_cv import _build_cv
        _build_cv(csrc_dir)
    else:
        raise RuntimeError(f"Unsupported lib: {name}")


def _load_or_build_lib_skip_if_error(name):
    try:
        return _load_or_build_lib(name)
    except Exception as e:
        logger.warning(
            f'Failed to load or JIT compile `{name}` extensions: \n{e}')

        
def _load_or_build_lib(name):
    if not _load_lib(name):
        _build_lib(name)
        return _load_lib(name)

if __name__ == "__main__":
    import fire
    fire.Fire({
        "build":  _build_lib
    })


import os
import shutil
import torch
import torchpipe
import fire
import subprocess
import signal
import pandas as pd
import matplotlib.pyplot as plt
from typing import List, Tuple, Dict
import numpy as np

def get_idle_gpu() -> int:
    """选择显存使用最小且利用率最低的GPU"""
    try:
        result = subprocess.run(
            ['nvidia-smi', '--query-gpu=index,memory.free,utilization.gpu',
             '--format=csv,noheader,nounits'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True
        )

        gpu_stats = []
        for line in result.stdout.strip().split('\n'):
            if not line:
                continue
            idx, mem_free, util = map(float, line.split(', '))
            gpu_stats.append({
                'id': int(idx),
                'mem_free': mem_free,
                'util': util
            })

        # 优先选择显存空闲最多的GPU，其次选择利用率最低的
        gpu_stats.sort(key=lambda x: (-x['mem_free'], x['util']))
        selected_gpu = gpu_stats[0]['id']
        print(
            f"Selected GPU {selected_gpu} - Free memory: {gpu_stats[0]['mem_free']} MB, Utilization: {gpu_stats[0]['util']}%")
        return selected_gpu

    except Exception as e:
        print(f"Error selecting GPU, using default 0: {str(e)}")
        return 0


def save_onnx(model: str, gpu_id: int):
    """在指定GPU上导出ONNX模型"""
    onnx_save_path = f"./{model}.onnx"
    if os.path.exists(onnx_save_path):
        return

    import timm
    assert model in timm.list_models(), f"Model {model} not in timm"

    # 设置GPU设备
    device = f"cuda:{gpu_id}"
    torch_model = timm.create_model(
        model, pretrained=False, exportable=True).eval()  # .to(device)

    torchpipe.utils.model_helper.onnx_export(
        torch_model, onnx_save_path, 224, 224)


def build_engine(model: str, batch_size: int, gpu_id: int):
    """在指定GPU上构建TensorRT引擎"""
    engine_path = f"./{model}_b{batch_size}.trt"
    if os.path.exists(engine_path):
        return engine_path

    build_cmd = [
        "trtexec",
        f"--onnx=./{model}.onnx",
        f"--saveEngine={engine_path}",
        f"--shapes=input:{batch_size}x3x224x224",
        "--fp16",
        f"--device={gpu_id}"  # 指定GPU
    ]
    subprocess.run(build_cmd, check=True)
    return engine_path


def run_benchmark(engine_path: str, batch_size: int, gpu_id: int, from_py: bool = False) -> Tuple[float, float, int]:
    """在指定GPU上运行基准测试"""
    # 准备GPU监控
    log_file = f"./gpu_log_{os.path.basename(engine_path)}.csv"

    # nvidia-smi -i=1 --query-gpu=utilization.gpu,memory.used --format=csv -l 1
    # nvidia-smi pmon   -s um -i 1
    # 启动nvidia-smi监控指定GPU
    with open(log_file, 'w') as f:
        proc_smi = subprocess.Popen(
            ['nvidia-smi',
             f'-i={gpu_id}',
             '--query-gpu=utilization.gpu,memory.used',
             '--format=csv', '-l', '1'],
            stdout=f, stderr=subprocess.PIPE
        )

    # 运行基准测试
    if from_py:
        my_env = os.environ.copy()
        my_env['CUDA_VISIBLE_DEVICES'] = str(gpu_id)
        run_cmd = [
            f"python",
            f'{engine_path}',
            f"--batch_size={batch_size}",
            f"--gpu_id={gpu_id}"  # 指定GPU
        ]
    else:
        my_env = os.environ.copy()
        run_cmd = [
            "trtexec",
            f"--loadEngine={engine_path}",
            f"--shapes=input:{batch_size}x3x224x224",
            "--iterations=100",
            f"--device={gpu_id}"  # 指定GPU
        ]

    try:
        result = subprocess.run(
            run_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=300,
            env=my_env
        )
    except subprocess.TimeoutExpired:
        print(f"Benchmark timed out for {engine_path}")
        proc_smi.terminate()
        return 0.0, 0.0, 0

    # 停止监控
    proc_smi.terminate()

    # 解析吞吐量
    throughput = 0.0
    for line in result.stdout.split('\n'):
        if "Throughput:" in line:
            try:
                throughput = float(line.split("Throughput:")[1].split('qps')[0].strip().rstrip())
            except (IndexError, ValueError):
                pass
            break

    # 解析GPU指标
    tp50_sm, max_mem = parse_gpu_log(log_file)

    return throughput, tp50_sm, max_mem


def parse_gpu_log(log_path: str) -> Tuple[float, int]:
    """解析GPU日志文件"""
    try:
        with open(log_path) as f:
            lines = f.readlines()[1:]  # 跳过标题行

        sm_vals = []
        mem_vals = []
        for line in lines:
            if not line.strip():
                continue
            parts = line.strip().split(',')
            if len(parts) < 2:
                continue
            try:
                sm_val = float(parts[0].split()[0])  # 提取利用率数值
                mem_val = int(parts[1].split()[0])   # 提取显存数值
                if sm_val < 10:
                    continue
                sm_vals.append(sm_val)
                mem_vals.append(mem_val)
            except (ValueError, IndexError):
                continue

        # tp50_sm = sum(sm_vals) / len(sm_vals) if sm_vals else 0.0
        tp50_sm = np.percentile(sm_vals, 50) if sm_vals else 0.0
        max_mem = max(mem_vals) if mem_vals else 0
        return tp50_sm, max_mem

    except Exception as e:
        print(f"Error parsing GPU log: {str(e)}")
        return 0.0, 0


# def plot_results(model: str, results: List[dict]):
#     """生成吞吐量对比图"""
#     df = pd.DataFrame(results)
#     if df.empty:
#         return

#     # 找出归一化batch size的吞吐量参考值
#     t32 = df[df['batch_size'] == 32]['throughput'].values
#     t32 = t32[0] if len(t32) > 0 else 0

#     plt.figure(figsize=(10, 6))
#     plt.plot(df['batch_size'], df['throughput'], 'o-', label='Throughput')

#     if t32 > 0:
#         # 绘制75%阈值线并找出最小batch size
#         threshold = 0.75 * t32
#         plt.axhline(threshold, color='r', linestyle='--', label='75% of T32')

#         valid_bs = df[(df['batch_size'] >= 1) &
#                       (df['batch_size'] <= 12) &
#                       (df['throughput'] >= threshold)]
#         if not valid_bs.empty:
#             min_bs = valid_bs['batch_size'].min()
#             plt.plot(min_bs, valid_bs[valid_bs['batch_size'] == min_bs]['throughput'].values[0],
#                      'ro', markersize=8, label=f'Min BS: {min_bs}')

#     plt.title(f'Model: {model}')
#     plt.xlabel('Batch Size')
#     plt.ylabel('Throughput (qps)')
#     plt.legend()
#     plt.grid(True)
#     plt.savefig(f"{model}_throughput.png")
#     plt.close()

def main(
    models: List[str] = ['seg.py','jpg_decode.py', 'mobilenetv2_100', 'resnet101',
                         'vit_base_patch16_siglip_224'],
    batch_range: Tuple[int, int] = (1, 16),
    norm_batch_size: int = 64
):
    # models: List[str] = ['resnet101', 'mobilenetv2_100',
    #                      'vit_base_patch16_siglip_224'],
    # 验证trtexec是否可用
    if isinstance(models, str):
        models = [models]
    print(f'models = {models}')
    assert shutil.which("trtexec") is not None, "trtexec not found in PATH"

    # 选择最空闲的GPU
    selected_gpu = get_idle_gpu()
    print(f"Using GPU: {selected_gpu} for all benchmarks")

    all_results = []

    for model in models:
        model_results = []
        try:
            if model.endswith('.py'):                
                tput, tp50_sm, max_mem = run_benchmark(
                    model, norm_batch_size, selected_gpu, from_py=True)
                tput *= norm_batch_size
                curr_result = {
                    'model': model,
                    'batch_size': norm_batch_size,
                    'throughput': int(tput),
                    'tp50_sm': tp50_sm,
                    'max_mem': max_mem
                }
                model_results.append(curr_result)
                t32 = tput

                # 基准测试batch size范围
                for bs in range(batch_range[0], batch_range[1] + 1):
                    tput, tp50_sm, max_mem = run_benchmark(
                        model, bs, selected_gpu, from_py=True)
                    tput *= bs
                    current = {
                        'model': model,
                        'batch_size': bs,
                        'throughput': int(tput),
                        'tp50_sm': tp50_sm,
                        'max_mem': max_mem
                    }
                    model_results.append(current)
                    print(f'\ncurrent: {current}\n')
            else:
                save_onnx(model, selected_gpu)

                # 基准测试归一化batch size
                engine_path = build_engine(model, norm_batch_size, selected_gpu)
                tput, tp50_sm, max_mem = run_benchmark(
                    engine_path, norm_batch_size, selected_gpu)
                tput *= norm_batch_size
                curr_result = {
                    'model': model,
                    'batch_size': norm_batch_size,
                    'throughput': int(tput),
                    'tp50_sm': tp50_sm,
                    'max_mem': max_mem
                }
                model_results.append(curr_result)
                t32 = tput

                # 基准测试batch size范围
                for bs in range(batch_range[0], batch_range[1] + 1):
                    engine_path = build_engine(model, bs, selected_gpu)
                    tput, tp50_sm, max_mem = run_benchmark(
                        engine_path, bs, selected_gpu)
                    tput *= bs
                    current = {
                        'model': model,
                        'batch_size': bs,
                        'throughput': int(tput),
                        'tp50_sm': tp50_sm,
                        'max_mem': max_mem
                    }
                    model_results.append(current)
                    print(f'\ncurrent: {current}\n')

            all_results.extend(model_results)
                
        except Exception as e:
            print(f"Error processing {model}: {str(e)}")
        
    # 保存综合结果
    csv_file = "benchmark_results.csv"
    file_exists = os.path.isfile(csv_file)
    pd.DataFrame(all_results).to_csv(
        csv_file, 
        mode='a',          # 追加模式
        index=False,       # 不写入行索引
        header=not file_exists  # 如果文件不存在才写入列名
    )
    print(f'result saved to {csv_file}')


if __name__ == "__main__":
    fire.Fire(main)




# # Copyright 2021-2026 NetEase.
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# #  http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.

# # throughput and lantecy test
# from __future__ import annotations

# version = "20241030"

# """! @package test-tools
# # update 0.0.1 2022-03-15 整理出基础版本。 by zsy
# # update 0.1.1 2022-03-16 增加一次发送多个请求。 by zsy
# # update 0.1.2 2022-03-18 在一次发送多个请求的情况下修复avg 时间的计算。 by zsy
# # update 0.1.3 2022-03-18 在一次发送多个请求的情况每个数据都随机。 by zsy
# # update 0.1.4 2022-03-18 文档， typo. by zsy
# # update 0.1.5 2022-03-24 整理文档，格式调整， typo, 增加MEAN. 实际上MEAN～=avg，
# #                         但是少了数据选取（choice）和结果打印等时间, 根据avg和MEAN的值的差距推断，这部分影响在千分之三以内；
# #                         可以限制最长边大小，需要手动取消# img=pre_resize(img) 的注释  by zsy
# # update 0.1.6 2022-05-26 增加输出当前进程cpu使用情况和内存使用情况中位数  by zsy
# # update 0.1.7 2022-06-17 可设置pid；pid不存在时提示并退出；cpu利用率过小时，不显示结果（大概率匹配到错误的进程）fix batch_size>1   by zsy wlc
# # update 0.1.8 2022-07-28  直接读取jpg binary， 不在预先解码（文件名后缀需要为 ".jpg", '.JPG', '.jpeg', '.JPEG'）
# # update 2022-09-08  增加callback， 用于接收和处理结果；此时推理函数返回类型需要是list类型；
# #                           当 total_number <=0 时，变为只跑一遍的模式；
# #                           图片数量过少时，不再宕机;   
# # update 2023-01-04  增加 PRELOAD_TYPE: yes no auto;                        
# # update 2023-04-21  重新整理成类，兼容通用自定义场景;                        
# # update 2023-06-02  对齐文档中的新的API，兼容广义场景（API仍处于测试阶段）;                        
# # update 2023-07-25  更改主要API为test_from_raw_file;                        
# # update 2023-08-17  增加测试结果的返回;                        
# # update 2023-11-09  增加gpu使用率中位数输出;                        
# # update 2024-04-24  恢复为单文件，并增加 ProcessAdaptor
# # update 2024-05-22  使用batch_size取代request_batch参数。但保持兼容性
# # update 2024-10-30  fix test_thrift_from_raw_file 读图过多耗时过长的问题
# # update 2024-10-30   移除batch_size参数 
# """


# # from curses import flash
# # import torch
# from timeit import default_timer as timer

# # import cv2
# import sys
# import random
# import os
# import threading
# import numpy as np

# # sys.path.insert(0, os.path.join("..", os.path.dirname(__file__)))
# # sys.path.insert(0, os.path.join(".", os.path.dirname(__file__)))

# # from ..device_tools import install_package

# from typing import List, Union, Callable, Tuple

# # from .Sampler import Sampler, RandomSampler, preload, SequentialV0Sampler, FileSampler

# from typing import List, Tuple
# import random, os


# class Sampler:
#     def __init__(self):
#         pass

#     def __call__(self, start_index: int) -> None:
#         raise NotImplementedError

#     def batchsize(self):
#         return 1


# class RandomSampler(Sampler):
#     def __init__(self, data_source: List, batch_size=1):
#         super().__init__()
#         self.data_source = data_source
#         self.batch_size = batch_size
#         assert batch_size > 0

#         assert 0 < len(data_source)
#         for i in range(batch_size):
#             if len(data_source) < batch_size:
#                 data_source.append(data_source[i])

#     def __call__(self, start_index: int):
#         data = random.sample(self.data_source, self.batch_size)
#         self.forward(data)

#     def forward(self, data: List):
#         raise RuntimeError("Requires users to implement this function")

#     def batchsize(self):
#         return self.batch_size


# class SequentialV0Sampler(Sampler):
#     def __init__(self, data: List, batch_size=1):
#         super().__init__()
#         self.data = data
#         self.batch_size = batch_size
#         assert len(data) >= batch_size

#     def __call__(self, start_index: int) -> None:
#         data = self.data[start_index : start_index + self.batch_size]
#         self.forward(data)

#     def batchsize(self):
#         return self.batch_size

#     def forward(self, data: List):
#         raise RuntimeError("Requires users to implement this function")


# class LoopSampler(Sampler):
#     def __init__(self, data: List, batch_size=1):
#         super().__init__()
#         self.data = data
#         self.batch_size = batch_size
#         assert len(data) >= batch_size
#         self.length = len(data) - batch_size + 1
#         for i in range(batch_size):
#             self.data.append(data[i])

#     def __call__(self, start_index: int) -> None:
#         start_index = start_index % (self.length)
#         data = self.data[start_index : start_index + self.batch_size]
#         self.forward(data)

#     def batchsize(self):
#         return self.batch_size

#     def forward(self, data: List):
#         raise RuntimeError("Requires users to implement this function")


# class FileSampler(LoopSampler):
#     def __init__(self, data: List, batch_size=1):
#         super().__init__(data, batch_size)
#         self.local_result = {}

#     def forward(self, data: List):
#         raw_bytes = []
#         for file_path in data:
#             with open(file_path, "rb") as f:
#                 raw_bytes.append((file_path, f.read()))
#         self.handle_data(raw_bytes)

#     def handle_data(self, raw_bytes):
#         raise RuntimeError("Requires users to implement this function")


# def preload(
#     file_dir, num_preload=1000, recursive=True, ext=[".jpg", ".JPG", ".jpeg", ".JPEG"]
# ) -> List[Tuple[str, bytes]]:
#     if not os.path.exists(file_dir):
#         raise RuntimeError(file_dir + " not exists")

#     list_images = []
#     result = []
#     if recursive:
#         for root, folders, filenames in os.walk(file_dir):
#             for filename in filenames:
#                 if os.path.splitext(filename)[-1] in ext:
#                     list_images.append(os.path.join(root, filename))
#     else:
#         list_images = [
#             x for x in os.listdir(file_dir) if os.path.splitext(x)[-1] in ext
#         ]
#         list_images = [os.path.join(file_dir, x) for x in list_images]

#     for file_path in list_images:
#         if num_preload <= 0:
#             file_bytes = None
#         else:
#             with open(file_path, "rb") as f:
#                 file_bytes = f.read()
#         result.append((file_path, file_bytes))
#         if len(result) == num_preload:
#             break
#     if len(result) == 0:
#         raise RuntimeError("find no vaild files. ext = " + ext)

#     return result


# from collections import namedtuple


# class TestParams:
#     def __init__(self, total_number, num_clients) -> None:
#         self.lock = threading.Lock()
#         self.result = {}
#         self.finish_condition = threading.Condition()

#         assert total_number >= 0
#         self.total_number = total_number
#         self.num_clients = num_clients


# # TestParams = namedtuple("TestParams", ["lock", "total_number", "num_clients", "finish_condition", "result"])
# # LocalResult = namedtuple("LocalResult", ["latency", "start_time", "end_time"])
# class LocalResult:
#     def __init__(self) -> None:
#         self.latency = []


# # def pre_resize(im, max_size=640.0):
# #     h, w, _ = im.shape
# #     if max(h, w) <= max_size:
# #         return im
# #     ratio = max_size / max(h, w)

# #     resized_w = int(ratio * w)
# #     resized_h = int(ratio * h)
# #     im = cv2.resize(im, dsize=(resized_w, resized_h))
# #     return im


# class InferThreadData(threading.Thread):
#     def __init__(self, index, test_params: TestParams, forward_class: Sampler) -> None:
#         threading.Thread.__init__(self, name=str(index))
#         self.params = test_params
#         self.batch_size = forward_class.batchsize()
#         self.index = index
#         self.forward_class = forward_class

#         self.num_clients = self.params.num_clients

#         self.start_index = 0
#         self.should_stop = False
#         self.update_local_data()
#         self.local_result = LocalResult()

#     def update_local_data(self):
#         with self.params.lock:
#             if self.params.total_number <= 0:
#                 self.should_stop = True
#                 return

#             if self.params.total_number >= self.batch_size:
#                 self.params.total_number -= self.batch_size
#                 self.start_index = self.params.total_number
#             else:
#                 self.params.total_number -= self.batch_size
#                 self.start_index = 0

#     def onFinish(self):
#         self.local_result.end_time = timer()
#         # while True:
#         should_wait = False
#         with self.params.lock:
#             self.params.num_clients -= 1
#             if self.params.num_clients != 0:
#                 should_wait = True

#         if should_wait:
#             with self.params.finish_condition:
#                 self.params.finish_condition.wait()
#         else:
#             with self.params.finish_condition:
#                 self.params.finish_condition.notify_all()
#         with self.params.lock:
#             self.params.result[self.index] = self.local_result
#             # print("self.onFinish()", self.params.result.keys(), flush=True)

#     def run(self):
#         self.local_result.start_time = timer()

#         try:
#             while not self.should_stop:
#                 self.forward(self.start_index)
#                 self.update_local_data()
#             self.onFinish()
#         except Exception as e:
#             import os

#             exc_type, exc_obj, exc_tb = sys.exc_info()
#             fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
#             print(exc_type, fname, exc_tb.tb_lineno)
#             print("subthread error: ", repr(e))
#             import os

#             os._exit(-1)

#         # self.forward_class.onFinish()

#     def forward(self, start_index):
#         start = timer()

#         self.forward_class(start_index)

#         result_time = timer() - start
#         self.local_result.latency.append((result_time, self.batch_size))
#         # return num_batch

#     def __del__(self):
#         pass

#     def warmup(self, num):
#         for i in range(num):
#             self.forward_class(i % (self.params.total_number))


# class GpuInfo(object):
#     def __init__(self, pid):
#         # 初始化
#         # nvml_lib = CDLL("/usr/lib/x86_64-linux-gnu/libnvidia-ml.so.1")
#         import pynvml  #  pip install  py3nvml # nvidia-ml-py3 pynvml

#         self.pynvml = pynvml
#         pynvml.nvmlInit()

#         self.need_record_index = -1  # 需要记录的进程PID

#         gpuDeviceCount = pynvml.nvmlDeviceGetCount()  # 获取Nvidia GPU块数
#         i = -1

#         CUDA_VISIBLE_DEVICES = os.environ.get("CUDA_VISIBLE_DEVICES", "0").split(",")
#         if len(CUDA_VISIBLE_DEVICES) == 1:
#             self.need_record_index = int(CUDA_VISIBLE_DEVICES[0])
#         else:
#             raise RuntimeError("CUDA_VISIBLE_DEVICES: only support single gpu")
#         # while self.need_record_index < 0:
#         #     i += 1
#         #     if i >= gpuDeviceCount:
#         #         assert(False, "node gpu not found")
#         #         break
#         #     print(i, gpuDeviceCount, type(torch.cuda.current_device()))
#         #     handle = pynvml.nvmlDeviceGetHandleByIndex(
#         #         i
#         #     )  # 获取GPU i的handle，后续通过handle来处理
#         #     # info = pynvml.nvmlDeviceGetMemoryInfo(handle)#通过handle获取GPU i的信息
#         #     ## gpu_memory_total = info.total #GPU i的总显存
#         #     # gpu_memory_used = info.used / NUM_EXPAND #转为MB单位
#         #     # all_gpu_used.append(gpu_memory_used) #添加进list

#         #     ###还可以直接针对pid的gpu消耗进行统计
#         #     info_list = pynvml.nvmlDeviceGetComputeRunningProcesses(
#         #         handle
#         #     )  # 获取所有GPU上正在运行的进程信息
#         #     # print(info_list)
#         #     # import pdb; pdb.set_trace()
#         #     info_list_len = len(info_list)
#         #     gpu_memory_used = 0
#         #     if info_list_len > 0:  # 0表示没有正在运行的进程
#         #         for info_i in info_list:
#         #             # print(info_i.pid, pid)
#         #             if info_i.pid == pid:  # 如果与需要记录的pid一致
#         #                 # gpu_memory_used += info_i.usedGpuMemory / NUM_EXPAND #统计某pid使用的总显存
#         #                 self.need_record_index = info_i
#         #                 break
#         self.handle = pynvml.nvmlDeviceGetHandleByIndex(self.need_record_index)
#         util = pynvml.nvmlDeviceGetUtilizationRates(self.handle).gpu

#     def get_pid_info(self):
#         util = self.pynvml.nvmlDeviceGetUtilizationRates(self.handle).gpu
#         return util

#     def get_gpu_device(self):
#         deviceCount = self.pynvml.nvmlDeviceGetCount()
#         gpu_list = []
#         for i in range(deviceCount):
#             handle = self.pynvml.nvmlDeviceGetHandleByIndex(i)
#             print("GPU", i, ":", self.pynvml.nvmlDeviceGetName(handle))
#             gpu_list.append(i)
#         return gpu_list

#     def get_free_rate(self, gpu_id):
#         handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_id)
#         info = pynvml.nvmlDeviceGetMemoryInfo(handle)
#         free_rate = int((info.free / info.total) * 100)
#         return free_rate

#     def get_gpu_info(self, gpu_id):
#         handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_id)
#         info = pynvml.nvmlDeviceGetMemoryInfo(handle)
#         M = 1024 * 1024
#         gpu_info = "id:{}  total:{}M free:{}M  used:{}M free_rate:{}%".format(
#             gpu_id,
#             info.total / M,
#             info.free / M,
#             info.used / M,
#             self.get_free_rate(gpu_id),
#         )
#         return gpu_info

#     def __del__(self):
#         # 最后要关闭管理工具
#         pass
#         # self.pynvml.nvmlShutdown()


# class ProcessAdaptor:
#     def __init__(self, class_def, args):
#         from multiprocessing import Process, Queue, Event

#         self.class_def = class_def
#         self.args = args

#         self.queue = Queue()
#         self.event = Event()
#         self.instance = Process(target=self.run)
#         self.alive = Event()

#         self.instance.start()

#     def forward(self, data):
#         self.queue.put(data)
#         # while not self.queue.empty():
#         self.event.wait()
#         self.event.clear()

#     def run(self):
#         self.target = self.class_def(self.args)
#         while not self.alive.is_set():
#             try:
#                 p = self.queue.get(block=True, timeout=2)
#                 if p is None:
#                     continue
#             except:
#                 continue
#             self.target.forward(p)
#             self.event.set()

#     def close(self):
#         self.alive.set()
#         self.instance.join()

#     @staticmethod
#     def close_all(clients):

#         from concurrent.futures import ThreadPoolExecutor

#         def close_client(client):
#             if hasattr(client, "close"):
#                 client.close()
#                 print("client closed")

#         # Assuming clients is a list of your client objects
#         with ThreadPoolExecutor() as executor:
#             executor.map(close_client, clients)


# # note 如果待测试函数有返回值，比如cuda上的tensor，有一定概率会copy到cpu并打印出来（初始概率下约打印10次，后续如果打印对象太大，则相应递减概率，但通常对性能影响小于千分之三
# class ResourceThread(threading.Thread):
#     def __init__(self, pid, result_list, my_event):
#         if pid == 0:
#             pid = os.getpid()
#             print(f"auto set --pid={pid}. Reset it if necessary")

#         threading.Thread.__init__(self, name="ResourceThread:" + str(pid))
#         import psutil

#         # self.p = psutil.Process(pid)
#         try:
#             self.p = psutil.Process(pid)
#         except:
#             print(f"pid {pid} not found")
#             exit(0)
#         print(f"Resource Monitor started: found {self.p.num_threads()} threads")
#         # print(psutil.pids())
#         # 'username',, 'status'
#         for proc in psutil.process_iter(["pid", "exe", "cmdline"]):
#             # if (proc.info["name"] in ["top", "cpptools-srv", "bash", "node", "cpptools", "sshd"]):
#             if pid != proc.info["pid"]:
#                 continue
#             print(proc.info)
#         # import time
#         # time.sleep(22)
#         self.result_list = result_list
#         self.my_event = my_event

#         self.gpu = None

#         try:
#             import pynvml

#             self.gpu = GpuInfo(pid)
#         except Exception as e:
#             print("gpu info not available: ", e)
#             self.gpu = None
#         # try:
#         #     self.gpu = GpuInfo(pid)
#         #     self.gpu.get_gpu_device()
#         # except Exception as e:
#         #     print(" pynvml :", e)

#     def get_cpu_mem(self):
#         return self.p.cpu_percent(), self.p.memory_percent()

#     def run(self):
#         import time

#         scale = 1
#         index = 0
#         while not self.my_event.wait(timeout=2):

#             index += 1
#             cpu_percent = self.p.cpu_percent()
#             mem_percent = self.p.memory_percent()
#             if cpu_percent > 0 and mem_percent > 0:
#                 if self.gpu and (index % scale == 0):
#                     util = self.gpu.get_pid_info()
#                     self.result_list.append((cpu_percent, mem_percent, util))
#                 else:
#                     self.result_list.append((cpu_percent, mem_percent, None))


# def test(sample: Union[Sampler, List[Sampler]], total_number=10000):
#     if isinstance(sample, list):
#         num_clients = len(sample)
#         assert num_clients > 0
#     elif isinstance(sample, Sampler):
#         sample = [sample]
#         num_clients = 1
#     else:
#         raise RuntimeError("must be Union[Sampler, List[Sampler]]")
#     assert total_number >= 0
#     test_params = TestParams(total_number, num_clients)

#     instance_threads = [
#         InferThreadData(i, test_params, sample[i]) for i in range(len(sample))
#     ]

#     warm_up_num = 20
#     print(f"Warm-up {warm_up_num} times for each client")
#     from concurrent.futures import ThreadPoolExecutor

#     with ThreadPoolExecutor(max_workers=num_clients) as t:
#         for thread_ in instance_threads:
#             t.submit(thread_.warmup, warm_up_num)

#     # torch.cuda.synchronize()
#     print("Warm-up finished", flush=True)

#     resource_result = []
#     resource_event = threading.Event()
#     resource_thread = ResourceThread(0, resource_result, resource_event)

#     resource_thread.daemon = True
#     resource_thread.start()

#     for i in instance_threads:
#         i.daemon = True

#     for i in instance_threads:
#         i.start()

#     for i in instance_threads:
#         i.join()
#     resource_event.set()

#     final_result = test_params.result

#     all_time = []
#     list_latency = []
#     for i in range(num_clients):
#         all_time.append(final_result[i].start_time)
#         all_time.append(final_result[i].end_time)
#         for pair_latency in final_result[i].latency:
#             list_latency += [pair_latency[0] * 1000] * pair_latency[1]
#     all_time.sort()

#     total_time = 1 * (all_time[len(all_time) - 1] - all_time[0])
#     resource_thread.join()

#     gpu_resource_result = []
#     try:
#         gpu_resource_result = [
#             x for x in list(zip(*resource_result))[2] if x is not None
#         ]
#     except:
#         pass

#     cpu_resource_result = list(zip(*resource_result))

#     try:
#         resource_result = np.array(cpu_resource_result)[:2, :].astype(np.float32)
#         cpu_ = int(10 * np.median(resource_result[0, :])) / 10
#         if cpu_ < 0.8 * 100:
#             cpu_ = 0
#     except:
#         cpu_ = "-"

#     try:
#         gpu_ = "-"
#         if gpu_resource_result:
#             gpu_ = int(10 * np.median(gpu_resource_result)) / 10
#             if gpu_ < 5:
#                 gpu_ = "-"
#     except Exception as e:
#         gpu_ = "-"
#         print("gpu_ error", e)

#     print("resource every 2s:")
#     print(resource_result)
#     print("")

#     if len(list_latency) + test_params.total_number != total_number:
#         print("len(list_latency): ", len(list_latency))
#         print(f"total_number: {total_number}")
#         print("test_params.total_number", test_params.total_number)
#         assert False
#     list_latency.sort()
#     while len(list_latency) < 50:
#         list_latency = list_latency * 2
#     length = len(list_latency)
#     tp50 = round(list_latency[length // 2], 2)
#     tp90 = round(list_latency[int(0.9 * length)], 2)
#     tp99 = round(list_latency[int(0.99 * length)], 2)
#     tp_1 = round(list_latency[-1], 2)
#     tp_2 = round(list_latency[-10], 2)
#     tp_3 = round(list_latency[-20], 2)
#     tp_4 = round(list_latency[-40], 2)
#     tp_5 = round(list_latency[-50], 2)
#     mean = round(sum(list_latency) / len(list_latency), 2)

#     qps = round(total_number / total_time, 2)
#     avg = round(1000 * num_clients / qps, 2)
    
#     print("------------------------------Summary------------------------------")
#     print(f"tool's version:: {version}")
#     print(f"num_clients:: {num_clients}")

#     print(f"total_number::   {total_number}")

#     print(f"throughput::     qps:  {qps},   [qps:=total_number/total_time]")
#     print(f"                 avg:  {avg} ms   [avg:=num_clients/qps]")

#     print(f"latency::        TP50: {tp50}   TP90: {tp90}   TP99:  {tp99} ")
#     print(
#         f"                 avg:  {mean}   -50,-40,-20,-10,-1: {tp_5},{tp_4},{tp_3},{tp_2},{tp_1} ms"
#     )
#     if cpu_ != "-":
#         print(f"cpu::            usage: {cpu_}%")
#     if gpu_ != "-":
#         print(f"gpu::            usage: {gpu_}%")
#     print(
#         "-------------------------------------------------------------------\n",
#         flush=True,
#     )

#     data = []
#     if False:
#         # x.field_names = ["Project",  "Value"]
#         data.append(["tool's version", version])
#         data.append(["num_clients", num_clients])
#         data.append(["total_number", total_number])

#         data.append(["throughput::qps", qps])
#         data.append(["throughput::avg", f"{avg}"])
#         data.append(["latency::TP50", f"{tp50}"])
#         data.append(["latency::TP90", f"{tp90}"])
#         data.append(["latency::TP99", f"{tp99}"])
#         data.append(["latency::avg", mean])
#         data.append(["-50,-40,-20,-10,-1", f"{tp_5},{tp_4},{tp_3},{tp_2},{tp_1}"])
#         try:
#             from prettytable import PrettyTable
#             import prettytable

#             x = PrettyTable()
#             x.field_names = ["Project", "Value"]
#             for item in data:
#                 x.add_row(item)
#             x.set_style(prettytable.MARKDOWN)
#             print("markdown style:\n\n", x, flush=True)
#         except:
#             print("if you need markdown style result, run: pip install PrettyTable")
#     result = {}
#     result["tool's version"] = version
#     result["num_clients"] = num_clients
#     result["total_number"] = total_number

#     result["throughput::qps"] = qps
#     result["throughput::avg"] = avg
#     result["latency::TP50"] = tp50
#     result["latency::TP90"] = tp90
#     result["latency::TP99"] = tp99
#     result["latency::avg"] = mean
#     result["-50"] = tp_5
#     result["-50"] = tp_4
#     result["-20"] = tp_3
#     result["-10"] = tp_2
#     result["-1"] = tp_1
#     result["cpu_usage"] = cpu_
#     result["gpu_usage"] = gpu_
#     return result

# def test_from_toml(
#     toml_path,
#     file_dir: str,
#     num_clients=10,
#     request_batch=1,
#     total_number=10000,
#     num_preload=1000,
#     recursive=True,
#     ext=[".jpg", ".JPG", ".jpeg", ".JPEG"],
# ):
#     import torchpipe

#     model = torchpipe.Pipe("vila.toml")

#     def run(inputs):
#         path_img, img = inputs[0]
#         input = {'data':img}
#         model(input)
#     return test_from_raw_file(run, file_dir, num_clients, 1, 1, total_number, num_preload, recursive, ext)

    
# def test_from_raw_file(
#     forward_function: Union[
#         Callable[[List[tuple[str, bytes]]]], List[Callable[[List[tuple[str, bytes]]]]]
#     ],
#     file_dir: str,
#     num_clients=10,
#     request_batch=1,
#     total_number=10000,
#     num_preload=1000,
#     recursive=True,
#     ext=[".jpg", ".JPG", ".jpeg", ".JPEG"],
# ):
#     """
#     This function is used to test the performance of a function.
#     It can be used to test the performance of a function that processes a single image, or a function that processes a batch of images.
#     """
#     data = preload(
#         file_dir=file_dir, recursive=recursive, num_preload=num_preload, ext=ext
#     )


#     print(f"file_dir = {file_dir}, num_clients = {num_clients}, request_batch = {request_batch}, total_number = {total_number}")
#     assert len(data) > 0
#     if num_preload <= 0:
#         total_number = len(data)

#     assert total_number > 0

#     if isinstance(forward_function, list):
#         assert len(forward_function) == num_clients
#     else:
#         forward_function = [forward_function] * num_clients

#     if num_preload > 0:
#         forwards = [RandomSampler(data, request_batch) for i in range(num_clients)]
#         for i in range(num_clients):
#             forwards[i].forward = forward_function[i]
#     else:
#         data = [x for x, _ in data]
#         forwards = [FileSampler(data, request_batch) for i in range(num_clients)]
#         for i in range(num_clients):
#             forwards[i].handle_data = forward_function[i]

#     return test(forwards, total_number)


# def test_function(
#     forward_function: Union[Callable, List[Callable]],
#     num_clients=10,
#     request_batch=1,
#     total_number=10000,
# ):
#     """
#     This function is used to test the performance of a function.
#     :param forward_function: a function or a list of functions.
#     :param num_clients: number of clients.
#     :param batch_size: batch size.
#     :param total_number: total number of data.
#     :return: None
#     """


#     class FunctionSampler:
#         def __init__(self, function, batchsize) -> None:
#             self.function = function
#             self.batch_size = batchsize

#         def __call__(self, start_index: int) -> None:
#             self.function()

#         def batchsize(self):
#             return self.batch_size

#     if isinstance(forward_function, list):
#         assert len(forward_function) == num_clients
#     else:
#         forward_function = [forward_function] * num_clients
#     forwards = [
#         FunctionSampler(forward_function[i], request_batch) for i in range(num_clients)
#     ]
#     return test(forwards, total_number)


# # import torchpipe

# class ProcessAdaptor:
#     def __init__(self, class_def, args):
#         from multiprocessing import Process, Queue, Event

#         self.class_def = class_def
#         self.args = args

#         self.queue = Queue()
#         self.event = Event()
#         self.instance = Process(target=self.run)
#         self.alive = Event()

#         self.instance.start()

#     def forward(self, data):
#         self.queue.put(data)
#         # while not self.queue.empty():
#         self.event.wait()
#         self.event.clear()

#     def run(self):
#         self.target = self.class_def(self.args)
#         while not self.alive.is_set():
#             try:
#                 p = self.queue.get(block=True, timeout=2)
#                 if p is None:
#                     continue
#             except:
#                 continue
#             self.target.forward(p)
#             self.event.set()

#     def close(self):
#         self.alive.set()
#         self.instance.join()

#     @staticmethod
#     def close_all(clients):

#         from concurrent.futures import ThreadPoolExecutor

#         def close_client(client):
#             if hasattr(client, "close"):
#                 client.close()

#         # Assuming clients is a list of your client objects
#         with ThreadPoolExecutor() as executor:
#             executor.map(close_client, clients)


# def parse_arguments(argv):
#     import argparse

#     parser = argparse.ArgumentParser()
#     parser.add_argument("--port", type=int, help="Port to listen.", default=8095)
#     parser.add_argument(
#         "--host", type=str, help="Host to run service", default="localhost"
#     )
#     parser.add_argument(
#         "--img_dir",
#         type=str,
#         help=f"img path. 预读取该目录以及子目录下至多1000张图片",
#         default="img/",
#     )
#     parser.add_argument("--batch_size", type=int, help="单次请求的数据量", default=1)
#     parser.add_argument("--num_clients", type=int, help="并发请求数", default=10)
#     return parser.parse_args(argv)


# if __name__ == "__main__":
#     total_number = 10000

#     args = parse_arguments(sys.argv[1:])

#     class ThriftInfer:
#         """wrapper for thrift's python API. You may need to re-implement this class."""

#         def __init__(self, host, port, batch_size) -> None:
#             """

#             :param host: ip
#             :type host: str
#             :param port: port
#             :type port: int
#             :param batch_size: size of sended data in batches.
#             :type batch_size: int
#             """
#             from serve import InferenceService
#             from serve.ttypes import InferenceParams

#             self.InferenceParams = InferenceParams

#             from thrift.transport import TSocket
#             from thrift.transport import TTransport
#             from thrift.protocol import TBinaryProtocol

#             self.transport = TSocket.TSocket(host, port)
#             self.transport = TTransport.TBufferedTransport(self.transport)
#             self.protocol = TBinaryProtocol.TBinaryProtocol(self.transport)

#             self.client = InferenceService.Client(self.protocol)

#             # Connect!
#             self.transport.open()
#             self.client.ping()
#             self.batch_size = batch_size

#         def infer(self, data):
#             """batch processing

#             :param data: batched data
#             :type data: List[(str, bytes)]
#             :return:
#             :rtype: Any
#             """
#             return self.client.infer_batch([self.InferenceParams(*x) for x in data])

#         def __del__(self):
#             self.transport.close()

#     def test_thrift_from_raw_file(
#         img_dir,
#         host="localhost",
#         port=8095,
#         num_clients=10,
#         request_batch = 1,
#         total_number=10000,
#     ):


#         instances_ = [ThriftInfer(host, port, request_batch) for i in range(num_clients)]

#         test_from_raw_file(
#             [x.infer for x in instances_],
#             img_dir,
#             num_clients,
#             request_batch,
#             -1,
#             total_number,
#         )

#     test_thrift_from_raw_file(
#         args.img_dir,
#         host=args.host,
#         port=args.port,
#         num_clients=args.num_clients,
#         request_batch=args.request_batch,
#         total_number=total_number,
#     )



# Copyright 2021-2026 NetEase.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from omniback.utils.system_path import system_include_dirs, system_library_dirs
import os
import sys
from ._cache_setting import get_cache_dir
import subprocess
import logging
logger = logging.getLogger(__name__)  # type: ignore


def is_system_exists_cv():
    exists_header = exists_lib = False
    for inc in system_include_dirs:
        if os.path.exists(os.path.join(inc, "opencv4/opencv2/core.hpp")):
            exists_header = True
            break
    for lib in system_library_dirs:
        if os.path.exists(os.path.join(lib, "libopencv_core.so")):
            exists_lib = True
            break
    return exists_lib and exists_header
def get_system_cv():
    exists_header = exists_lib = False
    for inc in system_include_dirs:
        if os.path.exists(os.path.join(inc, "opencv4/opencv2/core.hpp")):
            return os.path.join(inc, "opencv4/")
    raise RuntimeError('find no system opencv header')
    for lib in system_library_dirs:
        if os.path.exists(os.path.join(lib, "libopencv_core.so")):
            exists_lib = True
            break
    return exists_lib and exists_header

def can_use_cv_env():
    OPENCV_INCLUDE = os.environ.get("OPENCV_INCLUDE", None)
    OPENCV_LIB = os.environ.get("OPENCV_LIB", None)
    if OPENCV_INCLUDE and OPENCV_LIB:
        if not os.path.exists(OPENCV_INCLUDE, "opencv2/core.hpp"):
            raise RuntimeError(
                f"Wrong Env OPENCV_INCLUDE; can not find opencv header in dir {OPENCV_INCLUDE}")
        if not os.path.exists(OPENCV_LIB, "libopencv_core.so"):
            raise RuntimeError(
                f"Wrong Env OPENCV_LIB; can not find opencv libs in dir {OPENCV_LIB}")
        return True
    return False
    

def get_cv_include_lib_dir():
    # from env
    OPENCV_INCLUDE = os.environ.get("OPENCV_INCLUDE", None)
    OPENCV_LIB = os.environ.get("OPENCV_LIB", None)
    if OPENCV_INCLUDE and OPENCV_LIB:
        if not os.path.exists(OPENCV_INCLUDE, "opencv2/core.hpp"):
            raise RuntimeError(
                f"can not find opencv header in dir {OPENCV_INCLUDE}")
        if not os.path.exists(OPENCV_LIB, "libopencv_core.so"):
            raise RuntimeError(f"can not find opencv libs in dir {OPENCV_LIB}")
        return OPENCV_INCLUDE, OPENCV_LIB
    # from cache
    OPENCV_INCLUDE = OPENCV_LIB = None
    cache_header = os.path.join(get_cache_dir(), "opencv/include/opencv4/")
    cache_lib = os.path.join(get_cache_dir(), "opencv/lib/")
    possible_header_dirs = [cache_header]
    possible_lib_dirs = [cache_lib] +[os.path.join(get_cache_dir(), "opencv/lib64/")]
    for item in possible_header_dirs:
        if os.path.exists(os.path.join(item, "opencv2/core.hpp")):
            OPENCV_INCLUDE = item
            break
    for item in possible_lib_dirs:
        if os.path.exists(os.path.join(item, "libopencv_core.so")):
            OPENCV_LIB = item
            break
    if OPENCV_INCLUDE and OPENCV_LIB:
        return OPENCV_INCLUDE, OPENCV_LIB

    return None, None


def cache_cv_dir():
    import os
    import requests
    import zipfile
    import subprocess
    from pathlib import Path

    OPENCV_VERSION = "4.5.4"
    OPENCV_URL = f"https://codeload.github.com/opencv/opencv/zip/refs/tags/{OPENCV_VERSION}"
    OPENCV_ZIP = f"opencv-{OPENCV_VERSION}.zip"

    cache_dir = Path(get_cache_dir()) / "opencv"
    cache_dir.mkdir(parents=True, exist_ok=True)

    OPENCV_DIR = cache_dir / f"opencv-{OPENCV_VERSION}"
    cmake_lists_path = OPENCV_DIR / "CMakeLists.txt"

    # Download and extract if not exists
    if not cmake_lists_path.exists():
        zip_path = cache_dir / OPENCV_ZIP
        zip_path_cache = cache_dir / (OPENCV_ZIP+".cache")
        if not zip_path.exists():
            logger.warning(
                f"Downloading {OPENCV_URL} to {zip_path}")
            logger.warning(
                f'You may manully download it if it is too slow.')
            
            response = requests.get(OPENCV_URL, stream=True)
            response.raise_for_status()
            with open(zip_path_cache, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
            os.rename(zip_path_cache, zip_path)

        # Extract OpenCV
        print(f"Extracting {OPENCV_ZIP} to {cache_dir}...")
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(cache_dir)

    # Prepare build directory
    build_dir = OPENCV_DIR / "build"
    build_dir.mkdir(exist_ok=True)
    
    logger.warning(f"Building OpenCV {OPENCV_VERSION}... Skipping by providing envs OPENCV_INCLUDE and OPENCV_LIB.")
    logger.info(f"Building in {build_dir}")
    logger.warning(
        f'You can set envs OPENCV_INCLUDE and OPENCV_LIB to skip the downloading/building steps.')
    import omniback
    abi_flag = int(omniback.compiled_with_cxx11_abi())

    # CMake configuration
    cmake_args = [
        "cmake",
        f"-DCMAKE_CXX_FLAGS=-D_GLIBCXX_USE_CXX11_ABI={abi_flag}",
        "-DCMAKE_BUILD_TYPE=Release",
        "-DBUILD_WITH_DEBUG_INFO=OFF",
        f"-DCMAKE_INSTALL_PREFIX={cache_dir}",
        "-DINSTALL_C_EXAMPLES=OFF",
        "-DINSTALL_PYTHON_EXAMPLES=OFF",
        "-DENABLE_NEON=OFF",
        "-DBUILD_WEBP=OFF",
        "-DWITH_WEBP=OFF",
        "-DOPENCV_WEBP=OFF",
        "-DOPENCV_IO_ENABLE_WEBP=OFF",
        "-DHAVE_WEBP=OFF",
        "-DBUILD_ITT=OFF",
        "-DWITH_V4L=OFF",
        "-DWITH_QT=OFF",
        "-DWITH_OPENGL=OFF",
        "-DBUILD_opencv_dnn=OFF",
        "-DBUILD_opencv_java=OFF",
        "-DBUILD_opencv_python2=OFF",
        "-DBUILD_opencv_python3=OFF",
        "-DBUILD_NEW_PYTHON_SUPPORT=OFF",
        "-DBUILD_PYTHON_SUPPORT=OFF",
        "-DPYTHON_DEFAULT_EXECUTABLE=/usr/bin/python3",
        "-DBUILD_opencv_java_bindings_generator=OFF",
        "-DBUILD_opencv_python_bindings_generator=OFF",
        "-DBUILD_EXAMPLES=OFF",
        "-DWITH_OPENEXR=OFF",
        "-DWITH_JPEG=ON",
        "-DBUILD_JPEG=ON",
        "-DBUILD_JPEG_TURBO_DISABLE=OFF",
        "-DBUILD_DOCS=OFF",
        "-DBUILD_PERF_TESTS=OFF",
        "-DBUILD_TESTS=OFF",
        "-DWITH_PNG=OFF",
        "-DWITH_TIFF=OFF",
        "-DBUILD_opencv_apps=OFF",
        "-DBUILD_opencv_calib3d=OFF",
        "-DBUILD_opencv_contrib=OFF",
        "-DBUILD_opencv_features2d=OFF",
        "-DBUILD_opencv_flann=OFF",
        "-DBUILD_opencv_gapi=OFF",
        "-DWITH_CUDA=OFF",
        "-DWITH_CUDNN=OFF",
        "-DOPENCV_DNN_CUDA=OFF",
        "-DENABLE_FAST_MATH=1",
        "-DWITH_CUBLAS=0",
        "-DBUILD_opencv_gpu=OFF",
        "-DBUILD_opencv_ml=OFF",
        "-DBUILD_opencv_nonfree=OFF",
        "-DBUILD_opencv_objdetect=OFF",
        "-DBUILD_opencv_photo=OFF",
        "-DBUILD_opencv_stitching=OFF",
        "-DBUILD_opencv_superres=OFF",
        "-DBUILD_opencv_ts=OFF",
        "-DBUILD_opencv_video=OFF",
        "-DBUILD_videoio_plugins=OFF",
        "-DBUILD_opencv_videostab=OFF",
        "-DWITH_IPP=ON",
        "-DWITH_MKL=ON",
        "-DMKL_USE_TBB=ON",
        "-DWITH_TBB=ON",
        "-DBUILD_TBB=ON",
        "-DWITH_TURBOJPEG=ON",
        "-DWITH_LAPACK=ON",
        "-DWITH_BLAS=ON",
        str(OPENCV_DIR)  # Source directory
    ]

    # Build and install
    subprocess.run(cmake_args, cwd=build_dir, check=True, stdout=subprocess.DEVNULL,
                   stderr=subprocess.PIPE, 
                   text=True)
    subprocess.run(["make", "-j4"], cwd=build_dir, check=True)
    subprocess.run(["make", "install"], cwd=build_dir, check=True)

    # Return installation paths
    cache_header = cache_dir / "include/opencv4/"
    cache_lib = cache_dir / "lib/"
    if not cache_lib.exists():
        cache_lib = cache_dir / "lib64/"
        
    return str(cache_header), str(cache_lib)


def _build_cv(csrc_dir):
    # python -m omniback.utils.build_lib --no-torch --source-dirs csrc/mat_torch/ --include-dirs csrc/ /usr/local/include/opencv4/ --ldflags "-lopencv_core -lopencv_imgproc -lopencv_imgcodecs" --name torchpipe_opencv

    if not is_system_exists_cv() and not can_use_cv_env():
        cv_inc, cv_lib = get_cv_include_lib_dir()
        if cv_inc is None:
            cv_inc, cv_lib = cache_cv_dir()
        if cv_inc is None:
            raise RuntimeError(
                "OpenCV not found. Please specify its location using the "
                "OPENCV_INCLUDE and OPENCV_LIB environment variables."
            )
        os.environ["LD_LIBRARY_PATH"] = f"{cv_lib}:" + \
            os.environ.get("LD_LIBRARY_PATH", "")

        subprocess.run(
            [
                sys.executable,
                "-m",
                "omniback.utils.build_lib",
                "--source-dirs",
                os.path.join(csrc_dir, "csrc/mat_torch/"),
                "--include-dirs",
                os.path.join(csrc_dir, "csrc/"),
                f"{cv_inc}",
                "--no-torch",
                f"--ldflags=-L{cv_lib} -lopencv_core -lopencv_imgproc -lopencv_imgcodecs",
                "--name",
                "torchpipe_opencv"
            ],
            check=True,
            env={**os.environ, "EXAMPLE_ENV": "1"},
        )
    else:
        cv_inc = get_system_cv()
        subprocess.run(
            [
                sys.executable,
                "-m",
                "omniback.utils.build_lib",
                "--source-dirs",
                os.path.join(csrc_dir, "csrc/mat_torch/"),
                "--include-dirs",
                os.path.join(csrc_dir, "csrc/"),
                f"{cv_inc}",
                "--no-torch",
                f"--ldflags= -lopencv_core -lopencv_imgproc -lopencv_imgcodecs",
                "--name",
                "torchpipe_opencv"
            ],
            check=True,
            env={**os.environ, "EXAMPLE_ENV": "1"},
        )



# filePath: tests/test_decodeMat.py
import pytest
import omniback
import torchpipe
import requests
from io import BytesIO
from PIL import Image
import numpy as np
import torch
# Function to fetch image from URL
# def fetch_image(url):
#     try:
#         response = requests.get(url)
#         response.raise_for_status()  # Raise an exception for HTTP errors
#         return response.content
#     except requests.RequestException as e:
#         pytest.skip(f"Skipping test due to failure in fetching image: {e}")

# # Test function


def test_decodeMat():
    # Initialize the DecodeTensor model
    model = omniback.init("S[DecodeMat,Mat2Tensor]", {"color": "rgb"})

    img = np.ones((1, 1, 3), dtype=np.uint8)*5
    img[0, 0, 0] = 0
    img[0, 0, 1] = 0
    img[0, 0, 2] = 255
    img = Image.fromarray(img)
    img_byte_arr = BytesIO()
    img.save(img_byte_arr, format="JPEG")
    img_raw = img_byte_arr.getvalue()

    # Prepare input data
    input = {"data": img_raw}

    # Execute model inference
    model(input)
    result = input["result"]
    print(type(result.shape))
    if not isinstance(result, torch.Tensor):
        result = torch.from_dlpack(result)
    # Assert the shape of the output tensor
    assert result.shape == (1, 1, 3)
    # print(input["result"].shape)
    # print(input["result"][0,0,:])
    assert result[0, 0, -1] == 0


if __name__ == "__main__":
    import time
    # time.sleep(5)
    test_decodeMat()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Config utilities for finding paths to lib and headers."""

import argparse
import sys
from pathlib import Path

from . import libinfo


def _find_libdir() -> str:
    """Find the library directory for Omniback."""
    libomniback = Path(libinfo.find_libomniback())
    return str(libomniback.parent)


def __main__() -> None:  # noqa: PLR0912
    """Parse CLI args and print build and include configuration paths."""
    parser = argparse.ArgumentParser(
        description="Get various configuration information needed to compile with Omniback"
    )

    parser.add_argument("--includedir", action="store_true",
                        help="Print include directory")
    parser.add_argument(
        "--dlpack-includedir",
        action="store_true",
        help="Print dlpack include directory",
    )
    parser.add_argument("--cmakedir", action="store_true",
                        help="Print library directory")
    parser.add_argument("--sourcedir", action="store_true",
                        help="Print source directory")
    parser.add_argument("--libfiles", action="store_true",
                        help="Fully qualified library filenames")
    parser.add_argument("--libdir", action="store_true",
                        help="Print library directory")
    parser.add_argument("--libs", action="store_true",
                        help="Libraries to be linked")
    parser.add_argument("--cython-lib-path",
                        action="store_true", help="Print cython path")
    parser.add_argument("--cxxflags", action="store_true",
                        help="Print cxx flags")
    parser.add_argument("--cflags", action="store_true", help="Print c flags")
    parser.add_argument("--ldflags", action="store_true",
                        help="Print ld flags")

    args = parser.parse_args()

    # print help when no arguments are provided
    if len(sys.argv) == 1:
        parser.print_help()
        return

    if args.includedir:
        print(libinfo.find_include_path())
    if args.dlpack_includedir:
        print(libinfo.find_dlpack_include_path())
    if args.cmakedir:
        print(libinfo.find_cmake_path())
    if args.libdir:
        print(_find_libdir())
    if args.libfiles:
        if sys.platform.startswith("win32"):
            print(libinfo.find_windows_implib())
        else:
            print(libinfo.find_libomniback())
    if args.sourcedir:
        print(libinfo.find_source_path())
    if args.cython_lib_path:
        print(libinfo.find_cython_lib())
    if args.cxxflags:
        include_dir = libinfo.find_include_path()
        dlpack_include_dir = libinfo.find_dlpack_include_path()
        print(f"-I{include_dir} -I{dlpack_include_dir} -std=c++17")
    if args.cflags:
        include_dir = libinfo.find_include_path()
        dlpack_include_dir = libinfo.find_dlpack_include_path()
        print(f"-I{include_dir} -I{dlpack_include_dir}")
    if args.libs:
        if sys.platform.startswith("win32"):
            print(libinfo.find_windows_implib())
        else:
            print("-lomniback")
    if args.ldflags:
        if not sys.platform.startswith("win32"):
            print(f"-L{_find_libdir()}")


if __name__ == "__main__":
    __main__()



from __future__ import annotations

import os
import importlib

DIR = os.path.abspath(os.path.dirname(__file__))

def get_C_path() -> str:

    omniback_dir = os.path.join(DIR, "../")
    ext_suffix = importlib.machinery.EXTENSION_SUFFIXES[0]
    omniback_C_so = (os.path.join(omniback_dir, "omniback", f'_C{ext_suffix}'))
    return omniback_C_so


def get_root() -> str:
    return os.path.join(DIR, "../")


def get_includes() -> str:  # noqa: ARG001
    """
    Return the path to the omniback include directories.
    """
    in_directories = []
    installed_path = os.path.join(DIR, "include")
    if not os.path.exists(installed_path):
        installed_path = os.path.join(DIR, "../cpp/")
        assert os.path.exists(installed_path), installed_path
    re = [installed_path,  os.path.join(
        DIR, "./"),  os.path.join(DIR, "../3rdparty/spdlog/include/")]
    re += [os.path.join(DIR, "../")]
    return re
    source_path = os.path.join(os.path.dirname(DIR), "include")
    in_directories.append(installed_path if os.path.exists(
        installed_path) else source_path)

    curr = os.path.join(DIR, "../")
    assert os.path.exists(os.path.join(curr, "omniback", "csrc")), curr
    in_directories.append(curr)

    third = os.path.join(curr, "./3rdparty/spdlog/include/")
    assert os.path.exists(third)
    in_directories.append(third)
    return in_directories


def get_library_dir() -> str:  # noqa: ARG001
    lib_path = os.path.join(DIR, ".")
    so_path = os.path.join(lib_path, "libomniback.so")
    if not os.path.exists(so_path):
        lib_path = os.path.join(DIR, "../.setuptools-cmake-build/omniback")
        so_path = os.path.join(lib_path, "libomniback.so")
    if not os.path.exists(so_path):
        raise RuntimeError(f"{so_path} not exist")

    return lib_path


def get_cmake_dir() -> str:
    """
    Return the path to the omniback CMake module directory.
    """
    cmake_installed_path = os.path.join(DIR, "share", "cmake", "omniback")
    if os.path.exists(cmake_installed_path):
        return cmake_installed_path

    msg = "omniback not installed, installation required to access the CMake files"
    raise ImportError(msg)


def get_pkgconfig_dir() -> str:
    """
    Return the path to the omniback pkgconfig directory.
    """
    pkgconfig_installed_path = os.path.join(DIR, "share", "pkgconfig")
    if os.path.exists(pkgconfig_installed_path):
        return pkgconfig_installed_path

    msg = "omniback not installed, installation required to access the pkgconfig files"
    raise ImportError(msg)



import re
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from matplotlib.ticker import AutoLocator
from matplotlib.backends.backend_pdf import PdfPages


def parse_log_file(file_path):
    pattern = re.compile(
        r"model:(.*?),preprocessor:(.*?),tool's version:(.*?),num_clients:(.*?),"
        r"total_number:(.*?),throughput::qps:(.*?),throughput::avg:(.*?),"
        r"latency::TP50:(.*?),latency::TP90:(.*?),latency::TP99:(.*?),"
        r"latency::TP99\.9:(.*?),latency::TP99\.99:(.*?),latency::TP99\.999:(.*?),"
        r"latency::avg:(.*?),-50:(.*?),-20:(.*?),-10:(.*?),-1:(.*?),"
        r"cpu_usage:(.*?),gpu_usage:(.*?)$"
    )

    data = []
    with open(file_path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue

            match = pattern.match(line)
            if match:
                groups = match.groups()
                entry = {
                    'model': groups[0],
                    'preprocessor': groups[1],
                    'tool_version': groups[2],
                    'num_clients': int(groups[3]),
                    'total_requests': int(groups[4]),
                    'throughput_qps': float(groups[5]),
                    'throughput_avg': float(groups[6]),
                    '50': float(groups[7]),
                    '90': float(groups[8]),
                    '99': float(groups[9]),
                    '99.9': float(groups[10]),
                    '99.99': float(groups[11]),
                    '99.999': float(groups[12]),
                    'latency_avg': float(groups[13]),
                    'cpu_usage': float(groups[18]),
                    'gpu_usage': float(groups[19])
                }
                data.append(entry)
    data.sort(key=lambda entry: 'resnet' not in entry['model'])
    return pd.DataFrame(data)


def plot_percentile_comparison(df, output_file="latency_percentiles.pdf"):
    # 优化线条粗细设置
    plt.rcParams.update({
        "font.family": "serif",
        "font.serif": ["DejaVu Serif", "Times New Roman", "Times", "Palatino", "Bookman", "serif"],
        "font.size": 14,
        "axes.labelsize": 14,
        "axes.titlesize": 15,
        "legend.fontsize": 13,
        "xtick.labelsize": 13,
        "ytick.labelsize": 13,
        "lines.linewidth": 1.8,  # 减小线条宽度，使更协调
        "lines.markersize": 8,   # 减小标记尺寸
        "axes.linewidth": 1.0,   # 减细坐标轴线
        "grid.linewidth": 0.4,   # 减细网格线
        "savefig.dpi": 300,
        "savefig.bbox": "tight",
        "pdf.fonttype": 42
    })

    models = df['model'].unique().tolist()
    preprocessors = ['cpu', 'gpu']
    percentiles = [50, 90, 99, 99.9, 99.99, 99.999]
    percentile_labels = ['50', '90', '99', '99.9', '99.99', '99.999']
    # markers = ['o', 's', '^', 'D', 'v', '<']
    markers = ['']*10
    colors = {'cpu': '#1f77b4', 'gpu': '#d62728'}

    # 添加线条样式区分
    line_styles = {'cpu': '-', 'gpu': '--'}  # 实线 vs 虚线

    height_ratios = []
    model_ranges = []

    for model in models:
        model_df = df[df['model'] == model]
        all_vals = model_df[percentile_labels].values
        data_min = np.min(all_vals)
        data_max = np.max(all_vals)
        padding = (data_max - data_min) * 0.04 + 0.5
        y_min = max(0.1, data_min - padding)
        y_max = data_max + padding
        model_ranges.append((y_min, y_max))
        height_ratios.append(y_max - y_min)

    with PdfPages(output_file) as pdf:
        fig = plt.figure(figsize=(9, 8.5))

        gs = plt.GridSpec(len(models), 1, figure=fig,
                          height_ratios=height_ratios, hspace=0.0)
        axes = [fig.add_subplot(gs[i]) for i in range(len(models))]

        for i, model in enumerate(models):
            ax = axes[i]
            model_df = df[df['model'] == model]
            model_short = model
            y_min, y_max = model_ranges[i]

            for pp in preprocessors:
                pp_df = model_df[model_df['preprocessor'] == pp]
                if not pp_df.empty:
                    values = pp_df.iloc[0][percentile_labels].values
                    # 添加线条样式区分，减小标记尺寸
                    ax.plot(percentile_labels, values,
                            marker=markers[i],
                            markersize=7,      # 减小标记尺寸
                            linewidth=1.8,     # 减小线条宽度
                            linestyle=line_styles[pp],  # 添加线条样式
                            color=colors[pp],
                            label=f'{pp.upper()} Preprocessor',
                            zorder=3)

            # 优化标注框样式
            num_clients = model_df.iloc[0]['num_clients']
            ax.text(0.98, 0.15, f"Request Concurrency: {num_clients}",
                    transform=ax.transAxes,
                    fontsize=11,
                    color='#555555',
                    horizontalalignment='right',
                    verticalalignment='bottom',
                    bbox=dict(boxstyle='round,pad=0.3',
                              facecolor='white',
                              alpha=0.8,
                              edgecolor='#dddddd',  # 更浅的边框颜色
                              linewidth=0.8))       # 减细边框

            # 使用更柔和的网格线
            ax.grid(True, which="both", linestyle=':', alpha=0.5, zorder=1)
            ax.set_ylim(y_min, y_max)

            # 减细刻度线
            ax.tick_params(axis='y', which='both', length=4, width=0.8)
            ax.tick_params(axis='x', which='both', length=4, width=0.8)
            ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%d'))

            if i == len(models) - 1:
                ax.set_xlabel('Percentile', fontsize=16, labelpad=12)
            else:
                ax.set_xticklabels([])

            # 优化模型标识框
            ax.text(0.02, 0.85, f'({chr(97+i)}) {model_short}',
                    transform=ax.transAxes,
                    fontsize=16,
                    verticalalignment='top',
                    bbox=dict(boxstyle='round,pad=0.4',
                              fc='white',
                              ec='#888888',  # 更柔和的边框颜色
                              alpha=0.9,
                              linewidth=0.8))  # 减细边框

        fig.text(0.04, 0.5, 'Latency (ms)', va='center',
                 rotation='vertical', fontsize=16)

        # 优化图例位置和样式
        handles, labels = axes[0].get_legend_handles_labels()
        fig.legend(handles, labels, loc='lower right', ncol=1,
                   fontsize=13,  # 减小图例字体
                   bbox_to_anchor=(0.98, 0.03),
                   frameon=True,
                   framealpha=0.9,
                   edgecolor='#cccccc',  # 更柔和的边框颜色
                   title='',
                   title_fontsize=12)

        # 调整布局
        plt.tight_layout(pad=0.1, h_pad=0.0, w_pad=0.0,
                         rect=[0.06, 0.06, 0.95, 0.97])

        pdf.savefig(fig)
        plt.close()



if __name__ == "__main__":
    log_file = "../stability.log"
    output_pdf = "latency_percentiles.pdf"  # 更新输出文件名

    df = parse_log_file(log_file)
    plot_percentile_comparison(df, output_pdf)

    print(f"优化图表已保存至: {output_pdf}")



import os
import fire


def export_onnx(model_name="resnet101", output_dir="./"):
    """
    导出ONNX模型并转换为TensorRT引擎
    
    Args:
        model_name: 模型名称 (默认: resnet101)
        output_dir: 输出目录 (默认: ./)
    """
    import torch
    import timm

    # 确保输出目录存在
    os.makedirs(output_dir, exist_ok=True)

    if model_name in timm.list_models():
        model = timm.create_model(
            model_name, pretrained=False, exportable=True).eval()
    else:
        raise ValueError(f"model {model_name} not found in timm")

    # 创建示例输入
    x = torch.randn(1, 3, 224, 224)

    # 设置动态batch维度
    dynamic_axes = {
        'input': {0: 'batch_size'},  # 第0维（batch）为动态
        'output': {0: 'batch_size'}
    }

    # 导出ONNX模型
    onnx_path = f"{output_dir}/{model_name}.onnx"
    torch.onnx.export(
        model,
        x,
        onnx_path,
        export_params=True,
        opset_version=11,
        do_constant_folding=True,
        input_names=['input'],
        output_names=['output'],
        dynamic_axes=dynamic_axes
    )

    print(f"ONNX model saved to {onnx_path}")

    # 使用onnxslim简化模型
    simplified_onnx_path = onnx_path.replace('.onnx', '.onnx')
    os.system(f"onnxslim {onnx_path} {simplified_onnx_path}")
    print(f"Simplified ONNX model saved to {simplified_onnx_path}")
    return
    # 使用trtexec转换模型
    trt_cmd = (
        f"trtexec --onnx={simplified_onnx_path} "
        f"--saveEngine={output_dir}/model_repository/en/resnet_trt/1/model.plan "
        f"--explicitBatch "
        f"--minShapes=input:1x3x224x224 "
        f"--optShapes=input:{batch_size}x3x224x224 "
        f"--maxShapes=input:32x3x224x224 "  # 设置最大batch size为32
        f"--fp16 "
        f"--workspace=2048"
    )

    print(f"Running: {trt_cmd}")
    os.system(trt_cmd)
    print("TensorRT engine conversion completed!")


if __name__ == "__main__":
    fire.Fire(export_onnx)



import torch

import os
import importlib
import logging
def _get_extension_path(lib_name):

    lib_dir = os.path.dirname(__file__)
    assert os.name != "nt"

    loader_details = (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES)

    extfinder = importlib.machinery.FileFinder(lib_dir, loader_details)
    ext_specs = extfinder.find_spec(lib_name)
    if ext_specs is None:
        raise ImportError

    return ext_specs.origin

def _check_cuda_version():
    """
    Make sure that CUDA versions match between the pytorch install and torchpipe install
    """
    return True
    if not _HAS_OPS:
        return -1
    from torch.version import cuda as torch_version_cuda

    _version = torch.ops.torchvision._cuda_version()
    if _version != -1 and torch_version_cuda is not None:
        tv_version = str(_version)
        if int(tv_version) < 10000:
            tv_major = int(tv_version[0])
            tv_minor = int(tv_version[2])
        else:
            tv_major = int(tv_version[0:2])
            tv_minor = int(tv_version[3])
        t_version = torch_version_cuda.split(".")
        t_major = int(t_version[0])
        t_minor = int(t_version[1])
        if t_major != tv_major:
            raise RuntimeError(
                "Detected that PyTorch and torchvision were compiled with different CUDA major versions. "
                f"PyTorch has CUDA Version={t_major}.{t_minor} and torchvision has "
                f"CUDA Version={tv_major}.{tv_minor}. "
                "Please reinstall the torchvision that matches your PyTorch install."
            )
    return _version

import ctypes
def _load_library(lib_name):
    
    lib_path = _get_extension_path(lib_name)
    # try:
    #     ctypes.CDLL(lib_path)
    # except:
    logging.info(f"Loading {lib_path}")
    torch.ops.load_library(lib_path)


_check_cuda_version()


from modules import ClipModule, UNetModule, VaeModule, SafetyModule
import modules
import torch
import time
import json
import os
import fire
from copy import deepcopy
from tqdm import tqdm
from datetime import datetime
# from stable_diffusion_v1_5.stable_diffusion_pipeline import StableDiffusionPipeline
from typing import List, Dict, Any, Tuple
import numpy as np
import sys

torch.set_grad_enabled(False)

sys.path.insert(0, './')
assert os.environ["USE_TRT"]


class StableDiffusionPipeline():
    def __init__(self, config_path, **kwargs):
        super().__init__()
        fp = open(config_path, "r")
        config = json.load(fp)
        self.stream_module_list = []
        print(config.keys())
        torch.set_grad_enabled(False)

        self.modules = {'ClipModule': ClipModule,
                        'UNetModule': UNetModule,
                        'VaeModule': VaeModule,
                        'SafetyModule': SafetyModule}

        for key, value in config.items():
            self.modules[key] = self.modules[key](**value)
            # self.modules[key].deploy()
        self.stream_module_list = [self.modules['ClipModule'], self.modules['UNetModule'],
                                   self.modules['VaeModule'], self.modules['SafetyModule']]
        self.default_deploy()

    def default_deploy(self, **kwargs):
        for module in self.stream_module_list:
            module.deploy()


# 初始化管道
sd_config_file = "stable_diffusion_v1_5/config.json"
sd_pipeline = StableDiffusionPipeline(config_path=sd_config_file)

sd_modules = sd_pipeline.stream_module_list


def warmup_modules(
    modules: List,
    request_template: Dict[str, Any],
    warmup_iters: int = 5,
    progress_bar: bool = True
):
    """预热模块以稳定性能"""
    if progress_bar:
        print("\n🔥 Warming up modules...")
        pbar = tqdm(total=warmup_iters * len([1, 4, 8]) * len(modules),
                    desc="Warmup Progress", unit="op")

    for _ in range(warmup_iters):
        for batch_size in [1, 4, 8, 1]:  # 使用不同批次大小进行预热
            requests = [deepcopy(request_template) for _ in range(batch_size)]
            for module in modules:
                module.compute(requests)
                for req in requests:
                    for k, v in req.items():
                        if type(v) == torch.Tensor:
                            assert v.is_cpu == False
                torch.cuda.synchronize()
                if progress_bar:
                    pbar.update(1)

    if progress_bar:
        pbar.close()
        print("✅ Warmup completed")


def save_partial_results(results: Dict, output_path: str):
    """保存部分结果到文件"""
    with open(output_path, "w") as f:
        json.dump(results, f, ensure_ascii=True)


def profile_module_performance(
    modules: List,
    batch_sizes: List[int],
    request_template: Dict[str, Any],
    num_trials: int = 5,
    progress_bar: bool = True,
    output_path: str = None
) -> Dict[str, Dict[str, float]]:
    """性能测试函数，多次测量求平均，并增量保存结果"""
    latency_profiles = {type(m).__name__: {} for m in modules}

    # 创建进度条
    if progress_bar:
        total_ops = len(batch_sizes) * num_trials * len(modules)
        pbar = tqdm(total=total_ops, desc="Profiling Progress",
                    unit="batch-module")

    import sys
    sys.path.insert(
        0, './25Eurosys-NeuStream-AE/Diffusion/StableDiffusion/H100_SD_FP16_img512/')
    from test_set import prompt_list
    index = 0

    for batch_size in batch_sizes:
        # 存储每次试验的结果
        trial_results = {type(m).__name__: [] for m in modules}

        for trial in range(num_trials):
            # 创建请求批次（深拷贝避免引用问题）
            requests = []
            for i in range(batch_size):
                req = deepcopy(request_template)
                req['prompt'] = prompt_list[index % 100]
                index += 1
                requests.append(req)

            for module in modules:
                # 确保CUDA操作同步
                if type(module).__name__ == "UNetModule":
                    unet_loop_num = request_template["loop_num"].get(
                        "UNetModule", 50)
                else:
                    unet_loop_num = 1
                torch.cuda.current_stream().synchronize()

                # 精确测量执行时间
                start_time = time.perf_counter()
                for _ in range(unet_loop_num):
                    module.compute(requests)
                torch.cuda.current_stream().synchronize()

                # 计算耗时（毫秒）
                elapsed = (time.perf_counter() - start_time)

                # 如果是UNet模块，除以循环次数
                elapsed /= unet_loop_num

                trial_results[type(module).__name__].append(elapsed)

                # 更新进度条
                if progress_bar:
                    pbar.update(1)

        # 计算平均延迟
        for module_name, latencies in trial_results.items():
            if len(latencies) < 10:
                latencies += latencies
            avg_latency = np.mean(np.sort(latencies)[int(
                len(latencies)*0.1): -int(len(latencies)*0.1)])
            latency_profiles[module_name][str(batch_size)] = float(
                f"{avg_latency:.8f}")

        # 增量保存当前结果
        if output_path:
            new_latency_profiles = {}
            for module_name, v in latency_profiles.items():
                new_latency_profiles[module_name.replace(
                    'Module', '').lower()] = v
            save_partial_results(new_latency_profiles, output_path)

    new_latency_profiles = {}
    for module_name, v in latency_profiles.items():
        new_latency_profiles[module_name.replace('Module', '').lower()] = v

    if progress_bar:
        pbar.close()

    return new_latency_profiles


def generate_filename(image_size: int, config: Dict, max_batch: int) -> str:
    """生成包含配置信息的文件名"""
    import json
    import toml
    # with open('data/test_config.toml', 'r') as f:
    #     data = toml.load(f)
    # EXP_ID = os.getenv('EXP_ID')
    # json_p = data[EXP_ID]['latency_profile']

    return 'profiles/latency_profile.json'


def main(
    image_size: int = 256,
    min_batch: int = 1,
    max_batch: int = 48,
    num_trials: int = 5,
    warmup_iters: int = 3,
    config_path: str = "stable_diffusion_v1_5/config.json",
    output_dir: str = "profiles",
    progress_bar: bool = True
):
    """
    Stable Diffusion 模块性能分析工具
    
    参数:
    image_size (int): 图像尺寸 (默认: 256)
    min_batch (int): 最小批处理大小 (默认: 1)
    max_batch (int): 最大批处理大小 (默认: 40)
    num_trials (int): 每个批处理大小的测试次数 (默认: 5)
    warmup_iters (int): 预热迭代次数 (默认: 3)
    config_path (str): 配置文件路径 (默认: "stable_diffusion_v1_5/config.json")
    output_dir (str): 输出目录 (默认: "profiles")
    progress_bar (bool): 是否显示进度条 (默认: True)
    """
    print(f"\n🚀 Starting performance profiling with configuration:")
    print(f"  Image Size: {image_size}x{image_size}")
    print(f"  Batch Sizes: {min_batch} to {max_batch}")
    print(f"  Trials per Batch: {num_trials}")
    print(f"  Warmup Iterations: {warmup_iters}")
    print(f"  Config Path: {config_path}")

    # 创建输出目录
    os.makedirs(output_dir, exist_ok=True)

    # 配置和初始化
    # pipeline = initialize_pipeline(config_path)
    modules = sd_modules

    prompt = "a boy studying in Chinese University"
    # 请求模板
    REQUEST_TEMPLATE = {
        "prompt": prompt,
        "height": image_size,
        "width": image_size,
        "loop_num": {"UNetModule": 30},  # UNet循环50次
        "guidance_scale": 7.5,
        "seed": 81,
        "SLO": 10000,
        "loop_index": {"UNetModule": 0},
        "id": 1,
        "request_time": time.time()
    }

    # 性能测试参数
    BATCH_SIZES = list(range(min_batch, max_batch + 1))

    # 生成输出文件名
    filename = generate_filename(image_size, REQUEST_TEMPLATE, max_batch)
    output_path = filename  # os.path.join(output_dir, filename)

    # 预热模块
    warmup_modules(modules, REQUEST_TEMPLATE, warmup_iters, progress_bar)

    # 执行性能测试，并增量保存结果
    if progress_bar:
        print("\n📊 Starting performance profiling...")
    latency_results = profile_module_performance(
        modules=modules,
        batch_sizes=BATCH_SIZES,
        request_template=REQUEST_TEMPLATE,
        num_trials=num_trials,
        progress_bar=progress_bar,
        output_path=output_path
    )

    # 输出结果摘要
    print(f"\n✅ Profiling completed! Results saved to: {output_path}")
    print("\n📋 Summary of average latencies (s):")
    for module, data in latency_results.items():
        min_batch_latency = data.get(str(min_batch), "N/A")
        max_batch_latency = data.get(str(max_batch), "N/A")
        print(f"  {module}:")
        print(f"    Batch {min_batch}: {min_batch_latency}")
        print(f"    Batch {max_batch}: {max_batch_latency}")

    return output_path


if __name__ == "__main__":
    fire.Fire(main)



import pytest
from omniback import ffi
# from omniback.ffi import Dict as omniback_dict
import omniback as om

        

def test_any_containers():


    # Dictionary with float values
    test_dict = {"dd": 21.1, "ddd": 2.2}
    result_dict = om.Dict(test_dict)
    
    assert len(result_dict) == len(test_dict)
    for key, value in test_dict.items():
        assert result_dict[key] == value

def test_dict_operations():
    d = om.Dict({"key1": "value1"})
    d['key2'] = 2
    d['key3'] = "value3"
    
    assert d['key2'] == 2
    assert d['key3'] == "value3"
    
    d.pop('key2')
    assert 'key2' not in d
    assert len(d) == 2

def test_identity_backend():
    backend = om.create("Identity")
    backend.init({"param1": "value1"}).init({})
    
    # String input
    input_str = om.Dict({"data": "test_string"})
    backend(input_str)
    assert input_str['result'] == "test_string"
    assert backend.max() == 1
    
    # Integer input
    input_int = {"data": 42}
    backend(input_int)
    
    # Input with event
    input_event = {"data": 100, 'event': 1}
    input_event = om.Dict(input_event)
    backend(input_event)
    assert input_event['result'] == 100

if __name__ == "__main__":
    test_any_containers()



import pytest
import torch
import torchpipe
import os
import tempfile
import omniback

# Pipeline configuration strings
INIT_STR = "ModelLoadder[(.onnx)Onnx2Tensorrt,(.trt)LoadTensorrtEngine], TensorrtInferTensor"
FORWARD_STR = "CatSplit[S_v0[GpuTensor,CatTensor],S_v0[ContiguousTensor,TensorrtInferTensor,ProxyFromParam[post_processor]],SplitTensor]"
BACKEND_STR = f"StreamGuard[TensorrtTensor]"
# BACKEND_STR = "Identity"
class Conv(torch.nn.Module):
    def __init__(self):
        super(Conv, self).__init__()
        self.conv = torch.nn.Conv2d(3, 1, kernel_size=3, stride=2, padding=1)
    def forward(self, x):
        return self.conv(x)

def get_tmp_onnx(model: torch.nn.Module, input_shape: list) -> str:
    """
    Export PyTorch model to ONNX format
    
    Args:
        model: PyTorch model to export
        input_shape: Input tensor shape
    
    Returns:
        Path to temporary ONNX file
    """
    onnx_path = tempfile.mktemp(suffix=".onnx")
    model.eval()
    input_data = torch.randn(input_shape).cuda().half()
    torch.onnx.export(
        model, 
        input_data, 
        onnx_path,
        input_names=["input"], 
        output_names=["output"],
        dynamic_axes={"input": {0: "batch_size"}}
    )
    return onnx_path


def model_config():
    """Fixture to create model configuration and ONNX file"""
    config = {
        "model_type": "onnx",
        "post_processor": "Identity"
    }
    
    # Create temporary ONNX model
    torch_model = Conv().cuda().half()
    tmp_onnx = get_tmp_onnx(torch_model, [1, 3, 224, 224])
    config["model"] = tmp_onnx
    
    return config, torch_model, tmp_onnx
    
    # Cleanup temporary file
    os.remove(tmp_onnx)

def test_tensorrt_inference():
    """
    Test TensorRT inference pipeline
    
    Tests if the model correctly processes input tensor and produces expected output
    """
    
    config, torch_model, tmp_onnx = model_config()
    
    # Initialize model
    # model = omniback.init(BACKEND_STR, config)
    config["backend"] = BACKEND_STR
    model = omniback.init("Interpreter", config)
    
        
    
    # Prepare input data
    input_tensor = torch.ones((1, 3, 224, 224)).half()*10
    data = omniback.Dict({"data": input_tensor.cuda()})
    
    # Run inference
    # model(data)
    
    bench = omniback.init("Benchmark", {"num_clients": "4", "total_number": "10000"})
    bench.forward_with_dep([data]*100, model)
    
    q = omniback.default_queue()
    print(q.size())
    print(q)
    assert not q.empty()
    a=q.get(False)
    print(a)
    # import pdb;
    # pdb.set_trace()
    
    
    os.remove(tmp_onnx)
    return 
    # Verify results
    result = data['result']
    expected = torch_model(input_tensor.cuda())
    expected = expected.cuda()
    # print(result, expected)
    assert torch.allclose(result, expected, rtol=1e-2, atol=1e-2), "Model output does not match expected values"
    
if __name__ == "__main__":
    import time
    # time.sleep(5)
    # pytest.main(["-s", __file__])
    test_tensorrt_inference()



from pathlib import Path
from typing import Dict, Any, Union
import logging
import json
import os

from ._ffi_api import init, create, OmDict

logger = logging.getLogger("omniback")


try:
    import tomllib
except ImportError:
    import tomli as tomllib


def to_str(data) -> str:
    return str(data)


def to_dual_str(config: Dict[str, Dict[str, Any]]) -> Dict[str, Dict[str, str]]:
    """
    Convert a nested dictionary to a format where all values are strings.

    Args:
    config (Dict[str, Dict[str, Any]]): The input nested dictionary.

    Returns:
    Dict[str, Dict[str, str]]: The converted dictionary with all string values.

    Raises:
    TypeError: If the input is not a nested dictionary with the expected structure.
    """
    if not isinstance(config, dict):
        raise TypeError("Input must be a dictionary")

    re = {
        outer_key: {
            inner_key: str(int(inner_value)) if isinstance(
                inner_value, bool) else str(inner_value)
            for inner_key, inner_value in outer_value.items()
        }
        for outer_key, outer_value in config.items()
        if isinstance(outer_value, dict)
    }
    global_config = {outer_key: to_str(outer_value) for outer_key, outer_value in config.items()
                     if not isinstance(outer_value, dict)}
    re["global"] = global_config
    return re


def parse(file_path: Union[str, Path]) -> Dict[str, Dict[str, str]]:
    """
    Parse and convert TOML file content to string-based nested configuration.

    Args:
        file_path: Path to the TOML configuration file

    Returns:
        Nested dictionary with all values converted to strings

    Raises:
        FileNotFoundError: If specified file doesn't exist
        IOError: For file reading errors
        RuntimeError: For parsing/validation errors
    """
    # Convert to Path object for modern path handling
    path = Path(file_path)

    if not path.exists():
        raise FileNotFoundError(f"Config file not found: {file_path}")
    if not path.is_file():
        raise IOError(f"Path is not a file: {file_path}")

    try:
        with path.open('rb') as f:
            raw_config = tomllib.load(f)
    except tomllib.TOMLDecodeError as e:
        raise RuntimeError(f"TOML parsing failed: {e}") from e

    return to_dual_str(raw_config)


def log_structured_config(config: Dict[str, Any], title: str = "Configuration") -> None:
    """
    Log a structured configuration dictionary in a clean, formatted way.

    Args:
        config: Dictionary containing configuration data
        title: Optional title to display in the log header

    Returns:
        None
    """
    # Format the config as a pretty JSON string with indentation
    formatted_config = json.dumps(config, indent=2, ensure_ascii=False)

    # Add separators to make the log more readable
    logger.info("=" * 50)
    logger.info(f"{title}:")
    # Log each line separately to avoid excessively wide log entries
    for line in formatted_config.splitlines():
        logger.info(line)
    logger.info("=" * 50)


def pipe(config):
    if isinstance(config, str):
        if config.endswith('.toml'):
            assert os.path.exists(config), config
            return init_from_file(config)
    elif isinstance(config, dict):
        for k, v in config.items():
            if isinstance(v, dict):
                for k2, v2 in v.items():
                    if not isinstance(v, (bytes, str)):
                        config[k][k2] = str(v2)  # .encode("utf8")
            else:
                if not isinstance(v, (bytes, str)):
                    config[k] = str(v)  # .encode("utf8")
        print(config)
        return init("Interpreter", {}, OmDict({"config": (config)}), None)

    return init(config, {}, None, None)


def init_from_file(file_path: Union[str, Path]):
    """
    Initialize backend components from a TOML configuration file.

    Args:
        file_path: Path to the TOML configuration file

    Returns:
        Initialized Backend instance
    """
    logger.info(f"Loading configuration from: {file_path}")

    data = parse(file_path)

    # Log the configuration in a structured format
    log_structured_config(data, title="Configuration loaded")

    logger.info(f"Initializing interpreter with {len(data)} components")

    return init("Interpreter", {}, OmDict({"config": data}), None)
